# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the Access Control List (ACL) system.

The ACL systems is simple. It works by creating a many-to-many relationship
to the model that needs ACL functionality. This is implemented as a MixIn to
make it easy to annotate models to give them access to the ACL system.

The model has the following permissions: "read", "write" and "delete".
"""

import codecs
import json

import six

from flask_login import current_user
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import and_
from sqlalchemy import or_
from sqlalchemy import not_
from sqlalchemy import Unicode
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

from timesketch.models import BaseModel
from timesketch.models import db_session


class AccessControlEntry(object):
    """
    Access Control Entry database model. It has a user object (instance of
    timesketch.models.user.User) and a permission (read, write or delete).
    """

    @declared_attr
    def user_id(self):
        """Foreign key to a user model.

        Returns:
            A column (instance of sqlalchemy.Column)
        """
        return Column(Integer, ForeignKey('user.id'))

    @declared_attr
    def user(self):
        """A relationship to a user object.

        Returns:
            A relationship (instance of sqlalchemy.orm.relationship)
        """
        return relationship('User')

    @declared_attr
    def group_id(self):
        """Foreign key to a group model.

        Returns:
            A column (instance of sqlalchemy.Column)
        """
        return Column(Integer, ForeignKey('group.id'))

    @declared_attr
    def group(self):
        """A relationship to a group object.

        Returns:
            A relationship (instance of sqlalchemy.orm.relationship)
        """
        return relationship('Group')

    # Permission column (read, write or delete)
    permission = Column(Unicode(255))


class AccessControlMixin(object):
    """
    A MixIn for generating the necessary tables in the database and to make
    it accessible from the parent model object (the model object that uses this
    MixIn, i.e. the object that the ACL is added to).
    """

    @declared_attr
    def acl(self):
        """
        Generates the ACL tables and adds the attribute to the parent model
        object.

        Returns:
            A relationship to an ACE (timesketch.models.acl.AccessControlEntry)
        """
        self.AccessControlEntry = type(
            '%sAccessControlEntry' % self.__name__, (
                AccessControlEntry,
                BaseModel, ),
            dict(
                __tablename__='%s_accesscontrolentry' % self.__tablename__,
                parent_id=Column(Integer,
                                 ForeignKey('%s.id' % self.__tablename__)),
                parent=relationship(self), ))
        return relationship(self.AccessControlEntry)

    @classmethod
    def all_with_acl(cls, user=None):
        """
        Get all instances of the parent class that the user has read
        permission on. I.e enforce ACL permission check when fetching from
        the database.

        Args:
            user: A user (Instance of timesketch.models.user.User)

        Returns:
            An ACL base query (instance of timesketch.models.AclBaseQuery)
        """
        # If no user, assume the user that made the request.
        if not user:
            user = current_user

        # pylint: disable=singleton-comparison
        return cls.query.filter(
            or_(cls.AccessControlEntry.user == user,
                and_(cls.AccessControlEntry.user == None,
                     cls.AccessControlEntry.group == None),
                cls.AccessControlEntry.group_id.in_([
                    group.id for group in user.groups
                ])), cls.AccessControlEntry.permission == 'read',
            cls.AccessControlEntry.parent)

    def _get_ace(self, permission, user=None, group=None, check_group=True):
        """Get the specific access control entry for the user and permission.

        Args:
            permission: Permission as string (read, write or delete)
            user: A user (Instance of timesketch.models.user.User)
            group: A group (Instance of timesketch.models.user.Group)
            check_group: Check group permission, default is True.

        Returns:
            An ACE (instance of timesketch.models.acl.AccessControlEntry) or
            None if no ACE is found.
        """
        # If group is specified check if an ACE exist for it and return early.
        if group:
            return self.AccessControlEntry.query.filter_by(
                group=group, permission=permission, parent=self).all()

        # Check access for user.
        ace = self.AccessControlEntry.query.filter_by(
            user=user, group=None, permission=permission, parent=self).all()

        # If user doesn't have a direct ACE, check group permission.
        if (user and check_group) and not ace:
            group_intersection = set(user.groups) & set(self.groups)
            for _group in group_intersection:
                # Get group ACE with the requested permission.
                ace = self.AccessControlEntry.query.filter_by(
                    group=_group, permission=permission, parent=self).all()
                if ace:
                    return ace
        return ace

    @property
    def my_permissions(self):
        """Return a string with the permissions of the current user."""
        has_permissions = []

        permissions = ['read', 'write', 'delete']
        for permission in permissions:
            if self.has_permission(user=current_user, permission=permission):
                has_permissions.append(permission)

        if current_user.admin:
            has_permissions.append('admin')

        return json.dumps(has_permissions)

    @property
    def all_permissions(self):
        """Return a string with all object permissions."""
        return json.dumps(self.get_all_permissions())

    @property
    def groups(self):
        """List what groups have acess to this sketch.

        Returns:
            Set of groups (instance of timesketch.models.user.Group)
        """
        # pylint: disable=singleton-comparison
        group_aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.group == None),
            self.AccessControlEntry.parent == self).all()
        return set(ace.group for ace in group_aces)

    @property
    def is_public(self):
        """Determine if the ACL is open to everyone.

        Returns:
            An ACE (instance of timesketch.models.acl.AccessControlEntry) if the
            object is readable by everyone or None if the object is private.
        """
        return self._get_ace(permission='read', user=None, group=None)

    @property
    def collaborators(self):
        """Get list of users that have explicit read permission on the object.

        Returns:
            List of users (instances of timesketch.models.user.User)
        """
        # pylint: disable=singleton-comparison
        aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.user == self.user),
            not_(self.AccessControlEntry.user == None),
            self.AccessControlEntry.permission == 'read',
            self.AccessControlEntry.parent == self).all()
        return set(ace.user for ace in aces)

    def get_all_permissions(self):
        """Get a dict of all users/groups that have permission on the object.

        Returns:
            Dict with users/groups as the key and permissions as the value.
                Usernames are prepended by user/ and groups by groups/.
        """
        return_dict = {}

        # pylint: disable=singleton-comparison
        aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.user == None),
            self.AccessControlEntry.parent == self).all()

        for ace in aces:
            name = 'user/{0:s}'.format(ace.user.username)
            return_dict.setdefault(name, [])
            return_dict[name].append(ace.permission)

        group_aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.group == None),
            self.AccessControlEntry.parent == self).all()

        for ace in group_aces:
            name = 'group/{0:s}'.format(ace.group.name)
            return_dict.setdefault(name, [])
            return_dict[name].append(ace.permission)

        return return_dict

    def get_permissions(self, permission):
        """Get a dict of users and groups that have permission on the object.

        Args:
            permission (str): the permission string (read, write or delete)

        Returns:
            Dict with users, groups and an indication of whether the sketch
            is public. Values are a set of user objects
            (instances of timesketch.models.user.User) or group objects
            (instances of timesketch.models.user.Group)
        """
        return_dict = {}

        # pylint: disable=singleton-comparison
        aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.user == None),
            self.AccessControlEntry.permission == permission,
            self.AccessControlEntry.parent == self).all()

        return_dict['users'] = set(ace.user for ace in aces)

        group_aces = self.AccessControlEntry.query.filter(
            not_(self.AccessControlEntry.group == None),
            self.AccessControlEntry.permission == permission,
            self.AccessControlEntry.parent == self).all()

        return_dict['groups'] = set(ace.group for ace in group_aces)

        return_dict['is_public'] = self.is_public
        return return_dict

    def has_permission(self, user, permission):
        """Check if the user has a specific permission.

        Args:
            user: A user (Instance of timesketch.models.user.User)
            permission: Permission as string (read, write or delete)

        Returns:
            An ACE (instance of timesketch.models.acl.AccessControlEntry) if the
            user has the permission or None if the user do not have the
            permission.
        """
        public_ace = self.is_public
        if public_ace and permission == 'read':
            return public_ace
        if isinstance(permission, six.binary_type):
            permission = codecs.decode(permission, 'utf-8')
        return self._get_ace(permission=permission, user=user)

    def grant_permission(self, permission, user=None, group=None):
        """Grant permission to a user or group  with the specific permission.

        Args:
            permission: Permission as string (read, write or delete)
            user: A user (Instance of timesketch.models.user.User)
            group: A group (Instance of timesketch.models.user.Group)
        """
        # Grant permission to a group.
        if group and not self._get_ace(permission, group=group):
            self.acl.append(
                self.AccessControlEntry(permission=permission, group=group))
            db_session.commit()
            return

        # Grant permission to a user.
        if not self._get_ace(permission, user=user, check_group=False):
            self.acl.append(
                self.AccessControlEntry(permission=permission, user=user))
            db_session.commit()

    def revoke_permission(self, permission, user=None, group=None):
        """Revoke permission for user/group on the object.

        Args:
            permission: Permission as string (read, write or delete)
            user: A user (Instance of timesketch.models.user.User)
            group: A group (Instance of timesketch.models.user.Group)
        """
        # Revoke permission for a group.
        if group:
            group_ace = self._get_ace(permission=permission, group=group)
            if group_ace:
                for ace in group_ace:
                    self.acl.remove(ace)
                db_session.commit()
            return

        # Revoke permission for a user.
        user_ace = self._get_ace(permission=permission, user=user)
        if user_ace:
            for ace in user_ace:
                self.acl.remove(ace)
            db_session.commit()



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""URL routes for API resources."""

from __future__ import unicode_literals

from .resources.aggregation import AggregationGroupResource
from .resources.aggregation import AggregationGroupListResource
from .resources.aggregation import AggregationListResource
from .resources.aggregation import AggregationExploreResource
from .resources.aggregation import AggregationInfoResource
from .resources.aggregation import AggregationResource
from .resources.analysis import AnalysisResource
from .resources.analysis import AnalyzerRunResource
from .resources.analysis import AnalyzerSessionResource
from .resources.attribute import AttributeResource
from .resources.explore import ExploreResource
from .resources.event import EventResource
from .resources.event import EventAnnotationResource
from .resources.event import EventCreateResource
from .resources.event import EventTaggingResource
from .resources.event import CountEventsResource
from .resources.sketch import SketchResource
from .resources.sketch import SketchListResource
from .resources.archive import SketchArchiveResource
from .resources.information import VersionResource
from .resources.view import ViewResource
from .resources.view import ViewListResource
from .resources.searchtemplate import SearchTemplateResource
from .resources.searchtemplate import SearchTemplateListResource
from .resources.upload import UploadFileResource
from .resources.task import TaskResource
from .resources.story import StoryListResource
from .resources.story import StoryResource
from .resources.explore import QueryResource
from .resources.timeline import TimelineResource
from .resources.timeline import TimelineListResource
from .resources.searchindex import SearchIndexListResource
from .resources.searchindex import SearchIndexResource
from .resources.session import SessionResource
from .resources.user import UserListResource
from .resources.user import GroupListResource
from .resources.user import CollaboratorResource
from .resources.user import LoggedInUserResource
from .resources.graph import GraphListResource
from .resources.graph import GraphResource
from .resources.graph import GraphPluginListResource
from .resources.graph import GraphCacheResource


# Disable error for long line. Readability is more important than line
# length in this case.
# pylint: disable=line-too-long
API_ROUTES = [
    (SketchListResource, '/sketches/'),
    (SketchResource, '/sketches/<int:sketch_id>/'),
    (SketchArchiveResource, '/sketches/<int:sketch_id>/archive/'),
    (AnalysisResource, '/sketches/<int:sketch_id>/timelines/<int:timeline_id>/analysis/'),
    (AnalyzerRunResource, '/sketches/<int:sketch_id>/analyzer/'),
    (AnalyzerSessionResource, '/sketches/<int:sketch_id>/analyzer/sessions/<int:session_id>/'),
    (AggregationListResource, '/sketches/<int:sketch_id>/aggregation/'),
    (AggregationGroupResource, '/sketches/<int:sketch_id>/aggregation/group/<int:group_id>/'),
    (AggregationGroupListResource, '/sketches/<int:sketch_id>/aggregation/group/'),
    (AggregationExploreResource, '/sketches/<int:sketch_id>/aggregation/explore/'),
    (AggregationInfoResource, '/aggregation/info/'),
    (AggregationResource, '/sketches/<int:sketch_id>/aggregation/<int:aggregation_id>/'),
    (ExploreResource, '/sketches/<int:sketch_id>/explore/'),
    (EventResource, '/sketches/<int:sketch_id>/event/'),
    (EventTaggingResource, '/sketches/<int:sketch_id>/event/tagging/'),
    (EventAnnotationResource, '/sketches/<int:sketch_id>/event/annotate/'),
    (EventCreateResource, '/sketches/<int:sketch_id>/event/create/'),
    (ViewListResource, '/sketches/<int:sketch_id>/views/'),
    (AttributeResource, '/sketches/<int:sketch_id>/attribute/'),
    (ViewResource, '/sketches/<int:sketch_id>/views/<int:view_id>/'),
    (SearchTemplateListResource, '/searchtemplate/'),
    (SearchTemplateResource, '/searchtemplate/<int:searchtemplate_id>/'),
    (UploadFileResource, '/upload/'),
    (TaskResource, '/tasks/'),
    (StoryListResource, '/sketches/<int:sketch_id>/stories/'),
    (StoryResource, '/sketches/<int:sketch_id>/stories/<int:story_id>/'),
    (QueryResource, '/sketches/<int:sketch_id>/explore/query/'),
    (CountEventsResource, '/sketches/<int:sketch_id>/count/'),
    (TimelineListResource, '/sketches/<int:sketch_id>/timelines/'),
    (TimelineResource, '/sketches/<int:sketch_id>/timelines/<int:timeline_id>/'),
    (SearchIndexListResource, '/searchindices/'),
    (SearchIndexResource, '/searchindices/<int:searchindex_id>/'),
    (SessionResource, '/sketches/<int:sketch_id>/explore/sessions/<string:timeline_index>'),
    (UserListResource, '/users/'),
    (GroupListResource, '/groups/'),
    (CollaboratorResource, '/sketches/<int:sketch_id>/collaborators/'),
    (VersionResource, '/version/'),
    (LoggedInUserResource, '/users/me/'),
    (GraphListResource, '/sketches/<int:sketch_id>/graphs/'),
    (GraphResource, '/sketches/<int:sketch_id>/graphs/<int:graph_id>/'),
    (GraphPluginListResource, '/graphs/'),
    (GraphCacheResource, '/sketches/<int:sketch_id>/graph/'),
]



"""Tests for LoginPlugin."""
from __future__ import unicode_literals

from timesketch.lib.analyzers import login
from timesketch.lib.testlib import BaseTest


class TestLoginPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def test_parse_evtx_logon(self):
        """Test EVTX logon parsing."""
        string_list = []
        string_parsed = {}

        attributes = login.parse_evtx_logon_event(string_list, string_parsed)
        self.assertEqual(len(attributes), 0)

        string_list = [
            'S-1-5', '1', '2', '0x0034', 'S-1-5-5-4-54', 'secret_santa', '6',
            '7', '2', '9', '10', '11', '12', '13', '14', '15', '16', '17',
            '18', '19', '20', '21', '22']
        string_parsed = {
            'target_user_id': 'S-1-5-5-4-54',
            'target_user_name': 'secret_santa',
            'target_machine_name': 'rudolph'}

        attributes = login.parse_evtx_logon_event(string_list, string_parsed)
        self.assertEqual(len(attributes), 9)

        hostname = attributes.get('hostname', 'N/A')
        self.assertEqual(hostname, 'rudolph')

        session = attributes.get('session_id', 'N/A')
        self.assertEqual(session, string_list[3])

        logoff_type = attributes.get('logon_type', 'N/A')
        self.assertEqual(logoff_type, 'Interactive')

    def test_parse_evtx_logoff(self):
        """Test EVTX logoff parsing."""
        string_list = ['0', 'giljagaur', 'esja', '0x000342', '2']

        attributes = login.parse_evtx_logoff_event(string_list)

        self.assertEqual(len(attributes), 4)
        username = attributes.get('username', 'N/A')
        self.assertEqual(username, string_list[1])

        session = attributes.get('session_id', 'N/A')
        self.assertEqual(session, string_list[3])

        logoff_type = attributes.get('logon_type', 'N/A')
        self.assertEqual(logoff_type, 'Interactive')



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Task resources for version 1 of the Timesketch API."""

import datetime

from flask import jsonify
from flask import current_app
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.models.sketch import SearchIndex


class TaskResource(resources.ResourceMixin, Resource):
    """Resource to get information on celery task."""

    # pylint: disable=import-outside-toplevel
    def __init__(self):
        super(TaskResource, self).__init__()
        # pylint: disable=import-outside-toplevel
        from timesketch.app import create_celery_app
        self.celery = create_celery_app()

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        timeout_threshold_seconds = current_app.config.get(
            'CELERY_TASK_TIMEOUT', 7200)
        indices = SearchIndex.query.filter(
            SearchIndex.status.any(status='processing')).filter_by(
                user=current_user).all()
        schema = {'objects': [], 'meta': {}}
        for search_index in indices:
            if search_index.get_status.status == 'deleted':
                continue
            # pylint: disable=too-many-function-args
            celery_task = self.celery.AsyncResult(search_index.index_name)
            task = dict(
                task_id=celery_task.task_id,
                state=celery_task.state,
                successful=celery_task.successful(),
                name=search_index.name,
                result=False)
            if celery_task.state == 'SUCCESS':
                task['result'] = celery_task.result
            elif celery_task.state == 'PENDING':
                time_pending = (
                    search_index.updated_at - datetime.datetime.now())
                if time_pending.seconds > timeout_threshold_seconds:
                    search_index.set_status('timeout')
            schema['objects'].append(task)
        return jsonify(schema)



"""Tests for NtfsTimestompPlugin."""
from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers import ntfs_timestomp
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

class MockEvent(object):
    def __init__(self, source=None):
        if source:
            self.source = source
        else:
            self.source = {}
        self.label = ""

    def add_attributes(self, attributes):
        self.source = dict(self.source, **attributes)

    def add_label(self, label):
        self.label = label

    def commit(self):
        pass

class FileInfoTestCase(object):
    def __init__(self, name, std_info_timestamp, fn_timestamps,
                 expected_si_diffs, expected_fn_diffs, is_timestomp):
        self.name = name
        ref = 7357
        ts_desc = "TEST"
        std_event = MockEvent()
        file_names = [(MockEvent(), ts) for ts in fn_timestamps]

        self.file_info = ntfs_timestomp.FileInfo(ref, ts_desc, std_event,
                                                 std_info_timestamp, file_names)
        self.expected_fn_diffs = expected_fn_diffs
        self.expected_si_diffs = expected_si_diffs

        self.is_timestomp = is_timestomp

class TestNtfsTimestompPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    #def __init__(self, *args, **kwargs):
    #    super(TestNtfsTimestompPlugin, self).__init__(*args, **kwargs)

    @mock.patch(
        u'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_is_suspicious(self):
        """Test is_suspicious method."""
        analyzer = ntfs_timestomp.NtfsTimestompSketchPlugin('is_suspicious', 1)

        test_cases = [
            FileInfoTestCase(
                "no timestomp",
                1000000000000,
                [1000000000000],
                None,
                [None],
                False
            ),
            FileInfoTestCase(
                "multiple file_names and all of them are timestomped",
                0,
                [6000000000, 7000000000, 8000000000],
                [6000000000, 7000000000, 8000000000],
                [6000000000, 7000000000, 8000000000],
                True
            ),
            FileInfoTestCase(
                "one of the file_names matches exactly",
                0,
                [0, 7000000000, 8000000000],
                None,
                [None, None, None],
                False
            ),
            FileInfoTestCase(
                "file_name is within threshold",
                0,
                [analyzer.threshold, 7000000000, 8000000000],
                None,
                [None, None, None],
                False
            ),
            FileInfoTestCase(
                "file_name is within threshold",
                0,
                [600000000, 7000000000, 8000000000],
                None,
                [None, None, None],
                False
            )
        ]

        for tc in test_cases:
            ret = analyzer.is_suspicious(tc.file_info)

            std_diffs = tc.file_info.std_info_event.source.get('time_deltas')
            fn_diffs = [event.source.get('time_delta') for event, _ in
                        tc.file_info.file_names]

            self.assertEqual(ret, tc.is_timestomp)
            self.assertEqual(std_diffs, tc.expected_si_diffs)
            self.assertEqual(fn_diffs, tc.expected_fn_diffs)


    # Mock the Elasticsearch datastore.
    @mock.patch(
        u'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_analyzer(self):
        """Test analyzer."""
        # TODO: Write actual tests here.
        self.assertEqual(True, True)



"""Tests for SessionizerSketchPlugin."""

from __future__ import unicode_literals

import unittest
import mock

from timesketch.lib.analyzers.sessionizer import SessionizerSketchPlugin
from timesketch.lib.analyzers.base_sessionizer_test import _create_mock_event
from timesketch.lib.analyzers.base_sessionizer_test \
    import check_surrounding_events
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestSessionizerPlugin(BaseTest):
    """Tests the functionality of the sessionizing sketch analyzer, focusing
    on edge cases."""
    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_multiple_sessions(self):
        """Test multiple events, two of which are in the same session and
        one in a different session."""

        index = 'test_index'
        sketch_id = 1
        analyzer = SessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(datastore, 0, 3, time_diffs=[3000, 400000000])

        message = analyzer.run()
        self.assertEqual(
            message, 'Sessionizing completed, number of session created: 2')

        event1 = datastore.event_store['0']
        self.assertEqual(event1['_source']['session_id'], {'all_events': 1})
        event2 = datastore.event_store['101']
        self.assertEqual(event2['_source']['session_id'], {'all_events': 1})
        event3 = datastore.event_store['202']
        self.assertEqual(event3['_source']['session_id'], {'all_events': 2})
        check_surrounding_events(self, datastore, [202], 'all_events')

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_zero_time_diff(self):
        """Test events with no time difference between them are allocated
        correctly."""
        index = 'test_index'
        sketch_id = 1
        analyzer = SessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(datastore, 0, 2, time_diffs=[0])

        message = analyzer.run()
        self.assertEqual(
            message, 'Sessionizing completed, number of session created: 1')

        event1 = datastore.event_store['0']
        self.assertEqual(event1['_source']['session_id'], {'all_events': 1})
        event2 = datastore.event_store['100']
        self.assertEqual(event2['_source']['session_id'], {'all_events': 1})
        event3 = datastore.event_store['101']
        self.assertEqual(event3['_source']['session_id'], {'all_events': 1})

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_zero_events(self):
        """Test the behaviour of the sessionizer when given zero events."""
        index = 'test_index'
        sketch_id = 1
        analyzer = SessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(datastore, 0, 0)

        message = analyzer.run()
        self.assertEqual(
            message, 'Sessionizing completed, number of session created: 0')

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_one_event(self):
        """Test the behaviour of the sessionizer when given one event."""
        index = 'test_index'
        sketch_id = 1
        analyzer = SessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(datastore, 0, 1)

        message = analyzer.run()
        self.assertEqual(
            message, 'Sessionizing completed, number of session created: 1')

        event1 = datastore.event_store['0']
        self.assertEqual(event1['_source']['session_id'], {'all_events': 1})


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing analyzers."""

from __future__ import unicode_literals


class AnalysisManager(object):
    """The analyzer manager."""

    _class_registry = {}

    @classmethod
    def _build_dependencies(cls, analyzer_names):
        """Build a dependency list of analyzers.

        Args:
            analyzer_names (list): List of analyzer names.

        Returns:
            A list of sets of analyzer names. Each set represents
            one dependency group.

        Raises:
            KeyError: if class introduces circular dependencies.
        """
        dependency_tree = []
        dependencies = {}

        for analyzer_name in analyzer_names:
            analyzer_class = cls.get_analyzer(analyzer_name)
            dependencies[analyzer_name] = [
                x.lower() for x in analyzer_class.DEPENDENCIES]

        while dependencies:
            dependency_list = []
            for value in iter(dependencies.values()):
                dependency_list.extend(value)

            # Find items without a dependency.
            dependency_set = set(dependency_list) - set(dependencies.keys())
            dependency_set.update(
                name for name, dep in iter(dependencies.items()) if not dep)

            if not dependency_set:
                raise KeyError((
                    'Unable to build dependency tree, there is a circular '
                    'dependency somewhere'))

            dependency_tree.append(dependency_set)

            # Let's remove the entries already in the tree and start again.
            new_dependencies = {}
            for analyzer_name, analyzer_dependencies in dependencies.items():
                if not analyzer_dependencies:
                    continue
                new_dependencies[analyzer_name] = list(
                    set(analyzer_dependencies) - dependency_set)
            dependencies = new_dependencies

        return dependency_tree

    @classmethod
    def clear_registration(cls):
        """Clears all analyzer registration."""
        cls._class_ordering = []
        cls._class_registry = {}

    @classmethod
    def get_analyzers(cls, analyzer_names=None):
        """Retrieves the registered analyzers.

        Args:
            analyzer_names (list): List of analyzer names.

        Yields:
            tuple: containing:
                str: the uniquely identifying name of the analyzer
                type: the analyzer class.
        """
        # Get all analyzers if no specific ones have been requested.
        if not analyzer_names:
            analyzer_names = cls._class_registry.keys()

        completed_analyzers = set()
        for cluster in cls._build_dependencies(analyzer_names):
            for analyzer_name in cluster:
                if analyzer_name in completed_analyzers:
                    continue
                analyzer_class = cls.get_analyzer(analyzer_name)
                yield analyzer_name, analyzer_class
                completed_analyzers.add(analyzer_name)

    @classmethod
    def get_analyzer(cls, analyzer_name):
        """Retrieves a class object of a specific analyzer.

        Args:
            analyzer_name (str): name of the analyzer to retrieve.

        Returns:
            Analyzer class object.
        """
        return cls._class_registry[analyzer_name.lower()]

    @classmethod
    def register_analyzer(cls, analyzer_class):
        """Registers an analyzer class.

        The analyzer classes are identified by their lower case name.

        Args:
            analyzer_class (type): the analyzer class to register.

        Raises:
            KeyError: if class is already set for the corresponding name.
        """
        analyzer_name = analyzer_class.NAME.lower()
        if analyzer_name in cls._class_registry:
            raise KeyError('Class already set for name: {0:s}.'.format(
                analyzer_class.NAME))
        cls._class_registry[analyzer_name] = analyzer_class



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the user model."""

from __future__ import unicode_literals

from timesketch.lib.testlib import ModelBaseTest
from timesketch.models.user import Group
from timesketch.models.user import User


class UserModelTest(ModelBaseTest):
    """Tests the user model."""

    def test_user_model(self):
        """
        Test that the test user has the expected data stored in the
        database.
        """
        expected_result = frozenset([('name', 'test1'), ('username', 'test1')])
        self._test_db_object(expected_result=expected_result, model_cls=User)

    def test_set_password(self):
        """Test setting a password for the user."""
        self.assertIsNone(self.user1.set_password('test', rounds=4))

    def test_valid_password(self):
        """Test checking a valid password."""
        self.assertTrue(self.user1.check_password('test'))

    def test_invalid_password(self):
        """Test checking a invalid password."""
        self.assertFalse(self.user1.check_password('invalid password'))


class GroupModelTest(ModelBaseTest):
    """Tests the user model."""

    def test_group_model(self):
        """
        Test that the test group has the expected data stored in the
        database.
        """
        expected_result = frozenset([('name', 'test_group1'),
                                     ('display_name', 'test_group1'),
                                     ('description', 'test_group1')])
        self._test_db_object(expected_result=expected_result, model_cls=Group)

    def test_group_membership(self):
        """Test that user1 is member of the group."""
        self.assertTrue(self.user1 in self.group1.users)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph plugin for Windows services."""

from timesketch.lib.graphs.interface import BaseGraphPlugin
from timesketch.lib.graphs import manager


class WinServiceGraph(BaseGraphPlugin):
    """Graph plugin for Windows services."""

    NAME = 'WinService'
    DISPLAY_NAME = 'Windows services'

    def generate(self):
        """Generate the graph.

        Returns:
            Graph object instance.
        """
        query = 'event_identifier:7045'
        return_fields = ['computer_name', 'username', 'strings']

        events = self.event_stream(
            query_string=query, return_fields=return_fields)

        for event in events:
            computer_name = event['_source'].get('computer_name', 'UNKNOWN')
            username = event['_source'].get('username', 'UNKNOWN')
            event_strings = event['_source'].get('strings', [])

            # Skip event if we don't have enough data to build the graph.
            try:
                service_name = event_strings[0]
                image_path = event_strings[1]
                service_type = event_strings[2]
                start_type = event_strings[3]
            except IndexError:
                continue

            computer = self.graph.add_node(computer_name, {'type': 'computer'})
            user = self.graph.add_node(username, {'type': 'user'})
            service = self.graph.add_node(service_name, {
                'type': 'winservice', 'image_path': image_path
            })

            self.graph.add_edge(user, service, start_type, event)
            self.graph.add_edge(service, computer, service_type, event)

        self.graph.commit()

        return self.graph


manager.GraphManager.register_graph(WinServiceGraph)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for v1 of the Timesketch API."""
from __future__ import print_function
from __future__ import unicode_literals

import json
import mock

from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

from timesketch.api.v1.resources import ResourceMixin


class ResourceMixinTest(BaseTest):
    """Test ResourceMixin."""

    def test_to_json_empty_list(self):
        """Behavior of to_json when given an empty list."""
        response = ResourceMixin().to_json([])
        self.assertEqual(response.json, {
            'meta': {},
            'objects': [],
        })


class SketchListResourceTest(BaseTest):
    """Test SketchListResource."""
    resource_url = '/api/v1/sketches/'

    def test_sketch_list_resource(self):
        """Authenticated request to get list of sketches."""
        self.login()
        response = self.client.get(self.resource_url)
        self.assertEqual(len(response.json['objects']), 2)
        result = sorted(i['name'] for i in response.json['objects'])
        self.assertEqual(result, ['Test 1', 'Test 3'])
        self.assert200(response)

    def test_sketch_post_resource(self):
        """Authenticated request to create a sketch."""
        self.login()
        data = dict(name='test', description='test')
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)


class SketchResourceTest(BaseTest):
    """Test SketchResource."""
    resource_url = '/api/v1/sketches/1/'

    @mock.patch(
        'timesketch.api.v1.resources.ElasticsearchDataStore', MockDataStore)
    def test_sketch_resource(self):
        """Authenticated request to get a sketch."""
        self.login()
        response = self.client.get(self.resource_url)
        self.assertEqual(len(response.json['objects']), 1)
        self.assertEqual(len(response.json['objects'][0]['timelines']), 1)
        self.assertEqual(response.json['objects'][0]['name'], 'Test 1')
        self.assertIsInstance(response.json['meta']['emojis'], dict)
        self.assert200(response)

    def test_sketch_acl(self):
        """
        Authenticated request to get a sketch that the user do not have read
        permission on.
        """
        self.login()
        response = self.client.get('/api/v1/sketches/2/')
        self.assert403(response)


class ViewListResourceTest(BaseTest):
    """Test ViewListResource."""
    resource_url = '/api/v1/sketches/1/views/'

    def test_post_view_list_resource(self):
        """Authenticated request to create a view."""
        self.login()
        data = dict(
            name='test',
            new_searchtemplate=False,
            query='test',
            filter={},
            dsl={})
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        data['from_searchtemplate_id'] = 1
        response_with_searchtemplate = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)
        self.assertEqual(response_with_searchtemplate.status_code,
                         HTTP_STATUS_CODE_CREATED)


class ViewResourceTest(BaseTest):
    """Test ViewResource."""
    resource_url = '/api/v1/sketches/1/views/1/'

    def test_view_resource(self):
        """Authenticated request to get a view."""
        self.login()
        response = self.client.get(self.resource_url)
        self.assertEqual(len(response.json['objects']), 1)
        self.assertEqual(response.json['objects'][0]['name'], 'View 1')
        self.assert200(response)

    def test_post_view_resource(self):
        """Authenticated request to update a view."""
        self.login()
        data = dict(name='test', query='test', filter='{}')
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)

    def test_invalid_user_in_view(self):
        """Authenticated request to get a view for another user."""
        self.login()
        response = self.client.get('/api/v1/sketches/1/views/3/')
        self.assert403(response)

    def test_invalid_view(self):
        """Authenticated request to get a view for non existing view."""
        self.login()
        response = self.client.get('/api/v1/sketches/1/views/2/')
        self.assert404(response)


class SearchTemplateResourceTest(BaseTest):
    """Test Search template resource."""
    resource_url = '/api/v1/searchtemplate/1/'

    def test_searchtemplate_resource(self):
        """Authenticated request to get a search template."""
        self.login()
        response = self.client.get(self.resource_url)
        self.assertEqual(len(response.json['objects']), 1)
        self.assertEqual(response.json['objects'][0]['name'], 'template')
        self.assert200(response)

    def test_invalid_searchtemplate(self):
        """Authenticated request to get a non existing search template."""
        self.login()
        response = self.client.get('/api/v1/searchtemplate/2/')
        self.assert404(response)


class ExploreResourceTest(BaseTest):
    """Test ExploreResource."""
    resource_url = '/api/v1/sketches/1/explore/'
    expected_response = {
        'meta': {
            'timeline_names': {
                'test': 'Timeline 1'
            },
            'timeline_colors': {
                'test': 'FFFFFF'
            },
            'es_total_count': 1,
            'es_time': 5,
            'count_per_index': {},
            'scroll_id': ''
        },
        'objects': [{
            'sort': [1410593223000],
            '_type': 'plaso_event',
            '_source': {
                'timestamp': 1410593222543942,
                'message': 'Test event',
                'label': ['__ts_star'],
                'timestamp_desc': 'Content Modification Time',
                'datetime': '2014-09-13T07:27:03+00:00'
            },
            '_score': 'null',
            'selected': False,
            '_index': 'test',
            '_id': 'test'
        }]
    }

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_search(self):
        """Authenticated request to query the datastore."""
        self.login()
        data = dict(query='test', filter={})
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertDictEqual(response.json, self.expected_response)
        self.assert200(response)


class AggregationExploreResourceTest(BaseTest):
    """Test AggregationExploreResource."""

    resource_url = '/api/v1/sketches/1/aggregation/explore/'

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_heatmap_aggregation(self):
        """Authenticated request to get aggregation requests."""
        self.login()
        data = dict(aggregation_dsl='test')
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assert200(response)


class EventResourceTest(BaseTest):
    """Test EventResource."""
    resource_url = '/api/v1/sketches/1/event/'
    expected_response = {
        'objects': {
            'timestamp_desc': '',
            'timestamp': 1410895419859714,
            'label': '',
            'source_long': '',
            'source_short': '',
            'es_index': '',
            'es_id': '',
            'message': '',
            'datetime': '2014-09-16T19:23:40+00:00'
        }
    }

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_get_event(self):
        """Authenticated request to get an event from the datastore."""
        self.login()
        response = self.client.get(self.resource_url +
                                   '?searchindex_id=test&event_id=test')
        self.assertDictContainsSubset(self.expected_response, response.json)
        self.assert200(response)

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_invalid_index(self):
        """
        Authenticated request to get an event from the datastore, but in the
        wrong index.
        """
        self.login()
        response_400 = self.client.get(
            self.resource_url + '?searchindex_id=wrong_index&event_id=test')
        self.assert400(response_400)


class EventAnnotationResourceTest(BaseTest):
    """Test EventAnnotationResource."""
    resource_url = '/api/v1/sketches/1/event/annotate/'

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_post_annotate_resource(self):
        """Authenticated request to create an annotation."""
        self.login()
        for annotation_type in ['comment', 'label']:
            event = {
                '_type': 'test_event',
                '_index': 'test',
                '_id': 'test'
            }
            data = dict(
                annotation='test',
                annotation_type=annotation_type,
                events=[event])
            response = self.client.post(
                self.resource_url,
                data=json.dumps(data),
                content_type='application/json')
            self.assertIsInstance(response.json, dict)
            self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)

    def test_post_annotate_invalid_index_resource(self):
        """
        Authenticated request to create an annotation, but in the wrong index.
        """
        self.login()
        data = dict(
            annotation='test',
            annotation_type='comment',
            event_id='test',
            searchindex_id='invalid_searchindex')
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_BAD_REQUEST)


class SearchIndexResourceTest(BaseTest):
    """Test SearchIndexResource."""
    resource_url = '/api/v1/searchindices/'

    @mock.patch('timesketch.api.v1.resources.ElasticsearchDataStore',
                MockDataStore)
    def test_post_create_searchindex(self):
        """Authenticated request to create a searchindex."""
        self.login()
        data = dict(
            searchindex_name='test3', es_index_name='test3', public=False)
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data),
            content_type='application/json')
        self.assertIsInstance(response.json, dict)
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)


class TimelineListResourceTest(BaseTest):
    """Test TimelineList resource."""
    resource_url = '/api/v1/sketches/1/timelines/'

    def test_add_existing_timeline_resource(self):
        """Authenticated request to add a timeline to a sketch."""
        self.login()
        data = dict(timeline=1)
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_OK)

    def test_add_new_timeline_resource(self):
        """Authenticated request to add a timeline to a sketch."""
        self.login()
        data = dict(timeline=2)
        response = self.client.post(
            self.resource_url,
            data=json.dumps(data, ensure_ascii=False),
            content_type='application/json')
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_CREATED)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package handles setting up and providing the database connection."""

from __future__ import unicode_literals

from flask import abort
from flask_login import current_user
from flask_sqlalchemy import BaseQuery
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import func
from sqlalchemy import Integer

from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN

# The database session
engine = None
session_maker = sessionmaker()
db_session = scoped_session(session_maker)


def configure_engine(url):
    """Configure and setup the database session."""
    # These needs to be global because of the way Flask works.
    # pylint: disable=global-statement,global-variable-not-assigned
    # TODO: Can we wrap this in a class?
    global engine, session_maker, db_session
    engine = create_engine(url)
    # Set the query class to our own AclBaseQuery
    session_maker.configure(
        autocommit=False, autoflush=False, bind=engine, query_cls=AclBaseQuery)


def init_db():
    """Initialize the database based on the implemented models. This will setup
    the database on the first run.
    """
    BaseModel.metadata.create_all(bind=engine)
    BaseModel.query = db_session.query_property()
    return BaseModel


def drop_all():
    """Drop all tables in the database."""
    BaseModel.metadata.drop_all(bind=engine)


class AclBaseQuery(BaseQuery):
    """The query object used for models. It subclasses
    flask_sqlalchemy.BaseQuery and has the same methods as a SQLAlchemy Query
    as well.
    """

    def get_with_acl(self, model_id, user=current_user):
        """Get a database object with permission check enforced.

        Args:
            model_id: The integer ID of the model to get.
            user: User (instance of timesketch.models.user.User)

        Returns:
            A BaseQuery instance.
        """
        result_obj = self.get(model_id)
        if not result_obj:
            abort(HTTP_STATUS_CODE_NOT_FOUND)
        try:
            if result_obj.get_status.status == 'deleted':
                abort(HTTP_STATUS_CODE_NOT_FOUND)
        except AttributeError:
            pass
        if result_obj.is_public:
            return result_obj
        if not result_obj.has_permission(user=user, permission='read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN)
        return result_obj


@as_declarative()
class BaseModel(object):
    """Base class used for database models. It adds common model fields to all
    models classes that subclass it.
    """

    @declared_attr
    def __tablename__(self):
        return self.__name__.lower()

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime(), default=func.now())
    updated_at = Column(DateTime(), default=func.now(), onupdate=func.now())

    def update_modification_time(self):
        """Set updated_at field to current time."""
        self.updated_at = func.now()

    @classmethod
    def get_or_create(cls, **kwargs):
        """Get or create a database object.

        Returns:
            A model instance.
        """
        instance = cls.query.filter_by(**kwargs).first()
        if not instance:
            instance = cls(**kwargs)
            db_session.add(instance)
            db_session.commit()
        return instance



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for analyzers."""

from __future__ import unicode_literals

import datetime
import json
import logging
import os
import time
import traceback
import yaml

from flask import current_app

import pandas

from timesketch.lib import definitions
from timesketch.lib.datastores.elastic import ElasticsearchDataStore
from timesketch.models import db_session
from timesketch.models.sketch import Aggregation
from timesketch.models.sketch import Attribute
from timesketch.models.sketch import AttributeValue
from timesketch.models.sketch import AggregationGroup as SQLAggregationGroup
from timesketch.models.sketch import Event as SQLEvent
from timesketch.models.sketch import Sketch as SQLSketch
from timesketch.models.sketch import Story as SQLStory
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import View
from timesketch.models.sketch import Analysis


logger = logging.getLogger('timesketch.analyzers')


def _flush_datastore_decorator(func):
    """Decorator that flushes the bulk insert queue in the datastore."""
    def wrapper(self, *args, **kwargs):
        func_return = func(self, *args, **kwargs)
        self.datastore.flush_queued_events()
        return func_return
    return wrapper


def get_config_path(file_name):
    """Returns a path to a configuration file.

    Args:
        file_name: String that defines the config file name.

    Returns:
        The path to the configuration file or None if the file cannot be found.
    """
    path = os.path.join(os.path.sep, 'etc', 'timesketch', file_name)
    if os.path.isfile(path):
        return path

    path = os.path.join(
        os.path.dirname(__file__), '..', '..', '..', 'data', file_name)
    path = os.path.abspath(path)
    if os.path.isfile(path):
        return path

    return None


def get_yaml_config(file_name):
    """Return a dict parsed from a YAML file within the config directory.

    Args:
        file_name: String that defines the config file name.

    Returns:
        A dict with the parsed YAML content from the config file or
        an empty dict if the file is not found or YAML was unable
        to parse it.
    """
    path = get_config_path(file_name)
    if not path:
        return {}

    with open(path, 'r') as fh:
        try:
            return yaml.safe_load(fh)
        except yaml.parser.ParserError as exception:
            # pylint: disable=logging-format-interpolation
            logger.warning((
                'Unable to read in YAML config file, '
                'with error: {0!s}').format(exception))
            return {}


class Event(object):
    """Event object with helper methods.

    Attributes:
        datastore: Instance of ElasticsearchDatastore.
        sketch: Sketch ID or None if not provided.
        event_id: ID of the Event.
        event_type: Document type in Elasticsearch.
        index_name: The name of the Elasticsearch index.
        source: Source document from Elasticsearch.
    """
    def __init__(self, event, datastore, sketch=None):
        """Initialize Event object.

        Args:
            event: Dictionary of event from Elasticsearch.
            datastore: Instance of ElasticsearchDatastore.
            sketch: Optional instance of a Sketch object.

        Raises:
            KeyError if event dictionary is missing mandatory fields.
        """
        self.datastore = datastore
        self.sketch = sketch

        self.updated_event = {}

        try:
            self.event_id = event['_id']
            self.event_type = event['_type']
            self.index_name = event['_index']
            self.source = event.get('_source', None)
        except KeyError as e:
            raise KeyError('Malformed event: {0!s}'.format(e)) from e

    def _update(self, event):
        """Update event attributes to add.

        Args:
            event: Dictionary with new or updated values.
        """
        self.updated_event.update(event)

    def commit(self, event_dict=None):
        """Commit an event to Elasticsearch.

        Args:
            event_dict: (optional) Dictionary with updated event attributes.
            Defaults to self.updated_event.
        """
        if event_dict:
            event_to_commit = event_dict
        else:
            event_to_commit = self.updated_event

        if not event_to_commit:
            return

        self.datastore.import_event(
            self.index_name, self.event_type, event_id=self.event_id,
            event=event_to_commit, flush_interval=1)
        self.updated_event = {}

    def add_attributes(self, attributes):
        """Add key/values to an Event.

        Args:
            attributes: Dictionary with new or updated values to add.
        """
        self._update(attributes)

    def add_label(self, label, toggle=False):
        """Add label to the Event.

        Args:
            label: Label name.
            toggle: If True the label will be removed if it exists already.

        Raises: RuntimeError of sketch ID is missing.
        """
        if not self.sketch:
            raise RuntimeError('No sketch provided.')

        user_id = 0
        updated_event = self.datastore.set_label(
            self.index_name, self.event_id, self.event_type, self.sketch.id,
            user_id, label, toggle=toggle, single_update=False)
        self.commit(updated_event)

    def add_tags(self, tags):
        """Add tags to the Event.

        Args:
            tags: List of tags to add.
        """
        if not tags:
            return

        existing_tags = self.source.get('tag', [])
        new_tags = list(set().union(existing_tags, tags))
        updated_event_attribute = {'tag': new_tags}
        self._update(updated_event_attribute)

    def add_emojis(self, emojis):
        """Add emojis to the Event.

        Args:
            emojis: List of emojis to add (as unicode codepoints).
        """
        if not emojis:
            return

        existing_emoji_list = self.source.get('__ts_emojis', [])
        if not isinstance(existing_emoji_list, (list, tuple)):
            existing_emoji_list = []
        new_emoji_list = list(set().union(existing_emoji_list, emojis))
        updated_event_attribute = {'__ts_emojis': new_emoji_list}
        self._update(updated_event_attribute)

    def add_star(self):
        """Star event."""
        self.add_label(label='__ts_star')

    def add_comment(self, comment):
        """Add comment to event.

        Args:
            comment: Comment string.

        Raises:
            RuntimeError: if no sketch is present.
        """
        if not self.sketch:
            raise RuntimeError('No sketch provided.')

        searchindex = SearchIndex.query.filter_by(
            index_name=self.index_name).first()
        db_event = SQLEvent.get_or_create(
            sketch=self.sketch.sql_sketch, searchindex=searchindex,
            document_id=self.event_id)
        comment = SQLEvent.Comment(comment=comment, user=None)
        db_event.comments.append(comment)
        db_session.add(db_event)
        db_session.commit()
        self.add_label(label='__ts_comment')

    def add_human_readable(self, human_readable, analyzer_name, append=True):
        """Add a human readable string to event.

        Args:
            human_readable: human readable string.
            analyzer_name: string with the name of the analyzer that was
                used to generate the human_readable string.
            append: boolean defining whether the data should be appended
                or prepended to the human readable string, if it has already
                been defined. Defaults to True, and does nothing if
                human_readable is not defined.
        """
        existing_human_readable = self.source.get('human_readable', [])

        human_readable = '[{0:s}] {1:s}'.format(analyzer_name, human_readable)

        if human_readable in existing_human_readable:
            return

        if append:
            existing_human_readable.append(human_readable)
        else:
            existing_human_readable.insert(0, human_readable)

        updated_human_readable = {'human_readable': existing_human_readable}
        self._update(updated_human_readable)


class Sketch(object):
    """Sketch object with helper methods.

    Attributes:
        id: Sketch ID.
        sql_sketch: Instance of a SQLAlchemy Sketch object.
    """
    def __init__(self, sketch_id):
        """Initializes a Sketch object.

        Args:
            sketch_id: The Sketch ID.
        """
        self.id = sketch_id
        self.sql_sketch = SQLSketch.query.get(sketch_id)

        if not self.sql_sketch:
            raise RuntimeError('No such sketch')

    def add_aggregation(
            self, name, agg_name, agg_params, description='', view_id=None,
            chart_type=None, label=''):
        """Add aggregation to the sketch.

        Args:
            name: the name of the aggregation run.
            agg_name: the name of the aggregation class to run.
            agg_params: a dictionary of the parameters for the aggregation.
            description: description of the aggregation, visible in the UI,
                this is optional.
            view_id: optional ID of the view to attach the aggregation to.
            chart_type: string representing the chart type.
            label: string with a label to attach to the aggregation.
        """
        if not agg_name:
            raise ValueError('Aggregator name needs to be defined.')
        if not agg_params:
            raise ValueError('Aggregator parameters have to be defined.')

        if view_id:
            view = View.query.get(view_id)
        else:
            view = None

        if chart_type:
            agg_params['supported_charts'] = chart_type

        agg_json = json.dumps(agg_params)
        aggregation = Aggregation.get_or_create(
            name=name, description=description, agg_type=agg_name,
            parameters=agg_json, chart_type=chart_type, user=None,
            sketch=self.sql_sketch, view=view)

        if label:
            aggregation.add_label(label)
        db_session.add(aggregation)
        db_session.commit()
        return aggregation

    def add_aggregation_group(self, name, description='', view_id=None):
        """Add aggregation Group to the sketch.

        Args:
            name: the name of the aggregation run.
            description: optional description of the aggregation, visible in
                the UI.
            view_id: optional ID of the view to attach the aggregation to.
        """
        if not name:
            raise ValueError('Aggregator group name needs to be defined.')

        if view_id:
            view = View.query.get(view_id)
        else:
            view = None

        if not description:
            description = 'Created by an analyzer'

        aggregation_group = SQLAggregationGroup.get_or_create(
            name=name, description=description, user=None,
            sketch=self.sql_sketch, view=view)
        db_session.add(aggregation_group)
        db_session.commit()

        return AggregationGroup(aggregation_group)

    def add_view(self, view_name, analyzer_name, query_string=None,
                 query_dsl=None, query_filter=None, additional_fields=None):
        """Add saved view to the Sketch.

        Args:
            view_name: The name of the view.
            analyzer_name: The name of the analyzer.
            query_string: Elasticsearch query string.
            query_dsl: Dictionary with Elasticsearch DSL query.
            query_filter: Dictionary with Elasticsearch filters.
            additional_fields: A list with field names to include in the
                view output.

        Raises:
            ValueError: If both query_string an query_dsl are missing.

        Returns: An instance of a SQLAlchemy View object.
        """
        if not (query_string or query_dsl):
            raise ValueError('Both query_string and query_dsl are missing.')

        if not query_filter:
            query_filter = {'indices': '_all'}

        if additional_fields:
            query_filter['fields'] = [
                {'field': x.strip()} for x in additional_fields]

        description = 'analyzer: {0:s}'.format(analyzer_name)
        view = View.get_or_create(
            name=view_name, description=description, sketch=self.sql_sketch,
            user=None)
        view.description = description
        view.query_string = query_string
        view.query_filter = view.validate_filter(query_filter)
        view.query_dsl = query_dsl
        view.searchtemplate = None
        view.set_status(status='new')

        db_session.add(view)
        db_session.commit()
        return view

    def add_sketch_attribute(self, name, values, ontology='text'):
        """Add an attribute to the sketch.

        Args:
            name (str): The name of the attribute
            values (list): A list of strings, which contains the values of the
                attribute.
            ontology (str): Ontology of the attribute, matches with
                data/ontology.yaml.
        """
        # Check first whether the attribute already exists.
        attribute = Attribute.query.filter_by(name=name).first()

        if not attribute:
            attribute = Attribute(
                user=None,
                sketch=self.sql_sketch,
                name=name,
                ontology=ontology)
            db_session.add(attribute)
            db_session.commit()

        for value in values:
            attribute_value = AttributeValue(
                user=None,
                attribute=attribute,
                value=value)

            attribute.values.append(attribute_value)
            db_session.add(attribute_value)
            db_session.commit()

        db_session.add(attribute)
        db_session.commit()

    def add_story(self, title):
        """Add a story to the Sketch.

        Args:
            title: The name of the view.

        Raises:
            ValueError: If both query_string an query_dsl are missing.

        Returns:
            An instance of a Story object.
        """
        story = SQLStory.query.filter_by(
            title=title, sketch=self.sql_sketch, user=None).first()

        if story:
            return Story(story)

        story = SQLStory.get_or_create(
            title=title, content='[]', sketch=self.sql_sketch, user=None)
        db_session.add(story)
        db_session.commit()
        return Story(story)

    def get_all_indices(self):
        """List all indices in the Sketch.
        Returns:
            List of index names.
        """
        active_timelines = self.sql_sketch.active_timelines
        indices = [t.searchindex.index_name for t in active_timelines]
        return indices


class AggregationGroup(object):
    """Aggregation Group object with helper methods.

    Attributes:
        group (SQLAlchemy): Instance of a SQLAlchemy AggregationGroup object.
    """
    def __init__(self, aggregation_group):
        """Initializes the AggregationGroup object.

        Args:
            aggregation_group: SQLAlchemy AggregationGroup object.
        """
        self.group = aggregation_group
        self._orientation = 'layer'
        self._parameters = ''

    @property
    def id(self):
        """Returns the group ID."""
        return self.group.id

    @property
    def name(self):
        """Returns the group name."""
        return self.group.name

    def add_aggregation(self, aggregation_obj):
        """Add an aggregation object to the group.

        Args:
            aggregation_obj (Aggregation): the Aggregation objec.
        """
        self.group.aggregations.append(aggregation_obj)
        self.group.orientation = self._orientation
        db_session.add(aggregation_obj)
        db_session.add(self.group)
        db_session.commit()

    def commit(self):
        """Commit changes to DB."""
        self.group.orientation = self._orientation
        self.group.parameters = self._parameters
        db_session.add(self.group)
        db_session.commit()

    def set_orientation(self, orientation='layer'):
        """Sets how charts should be joined.

        Args:
            orienation: string that contains how they should be connected
                together, That is the chart orientation,  the options are:
                "layer", "horizontal" and "vertical". The default behavior
                is "layer".
        """
        orientation = orientation.lower()
        if orientation == 'layer' or orientation.starstwith('layer'):
            self._orientation = 'layer'
        elif orientation == 'horizontal' or orientation.startswith('hor'):
            self._orientation = 'horizontal'
        elif orientation == 'vertical' or orientation.startswith('ver'):
            self._orientation = 'vertical'
        self.commit()

    def set_vertical(self):
        """Sets the "orienation" to vertical."""
        self._orientation = 'vertical'
        self.commit()

    def set_horizontal(self):
        """Sets the "orientation" to horizontal."""
        self._orientation = 'horizontal'
        self.commit()

    def set_layered(self):
        """Sets the "orientation" to layer."""
        self._orientation = 'layer'
        self.commit()

    def set_parameters(self, parameters=None):
        """Sets the parameters for the aggregation group.

        Args:
            parameters: a JSON string or a dict with the parameters
                for the aggregation group.
        """
        if isinstance(parameters, dict):
            parameter_string = json.dumps(parameters)
        elif isinstance(parameters, str):
            parameter_string = parameters
        elif parameters is None:
            parameter_string = ''
        else:
            parameter_string = str(parameters)
        self._parameters = parameter_string
        self.commit()


class Story(object):
    """Story object with helper methods.

    Attributes:
        story (SQLAlchemy): Instance of a SQLAlchemy Story object.
    """
    def __init__(self, story):
        """Initializes a Story object.

        Args:
            story: SQLAlchemy Story object.
        """
        self.story = story

    @property
    def data(self):
        """Return back the content of the story object."""
        return json.loads(self.story.content)

    @staticmethod
    def _create_new_block():
        """Create a new block to be added to a Story.

        Returns:
            Dictionary with default block content.
        """
        block = {
            'componentName': '',
            'componentProps': {},
            'content': '',
            'edit': False,
            'showPanel': False,
            'isActive': False
        }
        return block

    def _commit(self, block):
        """Commit the Story to database.

        Args:
            block (dict): Block to add.
        """
        story_blocks = json.loads(self.story.content)
        story_blocks.append(block)
        self.story.content = json.dumps(story_blocks)
        db_session.add(self.story)
        db_session.commit()

    def add_text(self, text, skip_if_exists=False):
        """Add a text block to the Story.

        Args:
            text (str): text (markdown is supported) to add to the story.
            skip_if_exists (boolean): if set to True then the text
                will not be added if a block with this text already exists.
        """
        if skip_if_exists and self.data:
            for block in self.data:
                if not block:
                    continue
                if not isinstance(block, dict):
                    continue
                old_text = block.get('content')
                if not old_text:
                    continue
                if text == old_text:
                    return

        block = self._create_new_block()
        block['content'] = text
        self._commit(block)

    def add_aggregation(self, aggregation, agg_type=''):
        """Add a saved aggregation to the Story.

        Args:
            aggregation (Aggregation): Saved aggregation to add to the story.
            agg_type (str): string indicating the type of aggregation, can be:
                "table" or the name of the chart to be used, eg "barcharct",
                "hbarchart". Defaults to the value of supported_charts.
        """
        today = datetime.datetime.utcnow()
        block = self._create_new_block()
        parameter_dict = json.loads(aggregation.parameters)
        if agg_type:
            parameter_dict['supported_charts'] = agg_type
        else:
            agg_type = parameter_dict.get('supported_charts')
            # Neither agg_type nor supported_charts is set.
            if not agg_type:
                agg_type = 'table'
                parameter_dict['supported_charts'] = 'table'

        block['componentName'] = 'TsAggregationCompact'
        block['componentProps']['aggregation'] = {
            'agg_type': aggregation.agg_type,
            'id': aggregation.id,
            'name': aggregation.name,
            'chart_type': agg_type,
            'description': aggregation.description,
            'created_at': today.isoformat(),
            'updated_at': today.isoformat(),
            'parameters': json.dumps(parameter_dict),
            'user': {'username': None},
            }
        self._commit(block)

    def add_aggregation_group(self, aggregation_group):
        """Add an aggregation group to the Story.

        Args:
            aggregation_group (SQLAggregationGroup): Save aggregation group
                to add to the story.
        """
        if not isinstance(aggregation_group, AggregationGroup):
            return

        block = self._create_new_block()
        block['componentName'] = 'TsAggregationGroupCompact'
        block['componentProps']['aggregation_group'] = {
            'id': aggregation_group.id,
            'name': aggregation_group.name}
        self._commit(block)

    def add_view(self, view):
        """Add a saved view to the Story.

        Args:
            view (View): Saved view to add to the story.
        """
        block = self._create_new_block()
        block['componentName'] = 'TsViewEventList'
        block['componentProps']['view'] = {'id': view.id, 'name': view.name}
        self._commit(block)


class BaseIndexAnalyzer(object):
    """Base class for analyzers.

    Attributes:
        name: Analyzer name.
        index_name: Name if Elasticsearch index.
        datastore: Elasticsearch datastore client.
        sketch: Instance of Sketch object.
    """

    NAME = 'name'
    IS_SKETCH_ANALYZER = False

    # If this analyzer depends on another analyzer
    # it needs to be included in this frozenset by using
    # the indexer names.
    DEPENDENCIES = frozenset()

    # Used as hints to the frontend UI in order to render input forms.
    FORM_FIELDS = []

    # Configure how long an analyzer should run before the timeline
    # gets fully indexed.
    SECONDS_PER_WAIT = 10
    MAXIMUM_WAITS = 360

    def __init__(self, index_name):
        """Initialize the analyzer object.

        Args:
            index_name: Elasticsearch index name.
        """
        self.name = self.NAME
        self.index_name = index_name
        self.timeline_name = ''
        self.datastore = ElasticsearchDataStore(
            host=current_app.config['ELASTIC_HOST'],
            port=current_app.config['ELASTIC_PORT'])

        if not hasattr(self, 'sketch'):
            self.sketch = None

    def event_stream(
            self, query_string=None, query_filter=None, query_dsl=None,
            indices=None, return_fields=None, scroll=True):
        """Search ElasticSearch.

        Args:
            query_string: Query string.
            query_filter: Dictionary containing filters to apply.
            query_dsl: Dictionary containing Elasticsearch DSL query.
            indices: List of indices to query.
            return_fields: List of fields to return.
            scroll: Boolean determining whether we support scrolling searches
                or not. Defaults to True.

        Returns:
            Generator of Event objects.

        Raises:
            ValueError: if neither query_string or query_dsl is provided.
        """
        if not (query_string or query_dsl):
            raise ValueError('Both query_string and query_dsl are missing')

        if not query_filter:
            query_filter = {'indices': self.index_name}

        # If not provided we default to the message field as this will always
        # be present.
        if not return_fields:
            return_fields = ['message']

        # Make sure we always return tag, human_readable and emoji attributes.
        return_fields.extend(['tag', 'human_readable', '__ts_emojis'])
        return_fields = list(set(return_fields))

        if not indices:
            indices = [self.index_name]

        # Refresh the index to make sure it is searchable.
        for index in indices:
            self.datastore.client.indices.refresh(index=index)

        event_generator = self.datastore.search_stream(
            query_string=query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            indices=indices,
            return_fields=return_fields,
            enable_scroll=scroll,
        )
        for event in event_generator:
            yield Event(event, self.datastore, sketch=self.sketch)

    @_flush_datastore_decorator
    def run_wrapper(self, analysis_id):
        """A wrapper method to run the analyzer.

        This method is decorated to flush the bulk insert operation on the
        datastore. This makes sure that all events are indexed at exit.

        Returns:
            Return value of the run method.
        """
        analysis = Analysis.query.get(analysis_id)
        analysis.set_status('STARTED')

        timeline = analysis.timeline
        self.timeline_name = timeline.name
        searchindex = timeline.searchindex

        counter = 0
        while True:
            status = searchindex.get_status.status
            status = status.lower()
            if status == 'ready':
                break

            if status == 'fail':
                logger.error(
                    'Unable to run analyzer on a failed index ({0:s})'.format(
                        searchindex.index_name))
                return 'Failed'

            time.sleep(self.SECONDS_PER_WAIT)
            counter += 1
            if counter >= self.MAXIMUM_WAITS:
                logger.error(
                    'Indexing has taken too long time, aborting run of '
                    'analyzer')
                return 'Failed'
            # Refresh the searchindex object.
            db_session.refresh(searchindex)

        # Run the analyzer. Broad Exception catch to catch any error and store
        # the error in the DB for display in the UI.
        try:
            result = self.run()
            analysis.set_status('DONE')
        except Exception:  # pylint: disable=broad-except
            analysis.set_status('ERROR')
            result = traceback.format_exc()

        # Update database analysis object with result and status
        analysis.result = '{0:s}'.format(result)
        db_session.add(analysis)
        db_session.commit()

        return result

    def run(self):
        """Entry point for the analyzer."""
        raise NotImplementedError


class BaseSketchAnalyzer(BaseIndexAnalyzer):
    """Base class for sketch analyzers.

    Attributes:
        sketch: A Sketch instance.
    """

    NAME = 'name'
    IS_SKETCH_ANALYZER = True

    def __init__(self, index_name, sketch_id):
        """Initialize the analyzer object.

        Args:
            index_name: Elasticsearch index name.
            sketch_id: Sketch ID.
        """
        self.sketch = Sketch(sketch_id=sketch_id)
        super().__init__(index_name)

    def event_pandas(
            self, query_string=None, query_filter=None, query_dsl=None,
            indices=None, return_fields=None):
        """Search ElasticSearch.

        Args:
            query_string: Query string.
            query_filter: Dictionary containing filters to apply.
            query_dsl: Dictionary containing Elasticsearch DSL query.
            indices: List of indices to query.
            return_fields: List of fields to be included in the search results,
                if not included all fields will be included in the results.

        Returns:
            A python pandas object with all the events.

        Raises:
            ValueError: if neither query_string or query_dsl is provided.
        """
        if not (query_string or query_dsl):
            raise ValueError('Both query_string and query_dsl are missing')

        if not query_filter:
            query_filter = {'indices': self.index_name, 'size': 10000}

        if not indices:
            indices = [self.index_name]

        # Refresh the index to make sure it is searchable.
        for index in indices:
            self.datastore.client.indices.refresh(index=index)

        if return_fields:
            default_fields = definitions.DEFAULT_SOURCE_FIELDS
            return_fields.extend(default_fields)
            return_fields = list(set(return_fields))
            return_fields = ','.join(return_fields)

        results = self.datastore.search_stream(
            sketch_id=self.sketch.id,
            query_string=query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            indices=indices,
            return_fields=return_fields,
        )

        events = []
        for event in results:
            source = event.get('_source')
            source['_id'] = event.get('_id')
            source['_type'] = event.get('_type')
            source['_index'] = event.get('_index')
            events.append(source)

        return pandas.DataFrame(events)

    def run(self):
        """Entry point for the analyzer."""
        raise NotImplementedError



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing graphs."""


class GraphManager:
    """The graph manager."""

    _class_registry = {}

    @classmethod
    def get_graphs(cls):
        """Retrieves the registered graphs.

        Yields:
            tuple: containing:
                str: the uniquely identifying name of the graph
                type: the graph class.
        """
        for graph_name, graph_class in iter(cls._class_registry.items()):
            yield graph_name, graph_class

    @classmethod
    def get_graph(cls, graph_name):
        """Retrieves a class object of a specific graph.

        Args:
            graph_name (str): name of the graph to retrieve.

        Returns:
            Instance of Graph class object.

        Raises:
            KeyError: if the graph is not registered.
        """
        try:
            graph_class = cls._class_registry[graph_name.lower()]
        except KeyError:
            raise KeyError(f'No such graph type: {graph_name.lower()}')
        return graph_class

    @classmethod
    def register_graph(cls, graph_class):
        """Registers an graph class.

        The graph classes are identified by their lower case name.

        Args:
            graph_class (type): the graph class to register.

        Raises:
            KeyError: if class is already set for the corresponding name.
        """
        graph_name = graph_class.NAME.lower()
        if graph_name in cls._class_registry:
            raise KeyError(f'Class already set for name: {graph_class.NAME}.')
        cls._class_registry[graph_name] = graph_class

    @classmethod
    def clear_registration(cls):
        """Clears all graph registrations."""
        cls._class_registry = {}



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Entry point for the application."""


# TODO: Remove once a new plaso release comes out.
def create_app(config=None):
    """Create the Flask app instance that is used throughout the application.

    Args:
        config: Path to configuration file as a string or an object with config
        directives.

    Returns:
        Application object (instance of flask.Flask).
    """
    # pylint: disable=import-outside-toplevel
    from timesketch import app
    return app.create_app(config)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Error classes."""

from __future__ import unicode_literals

from flask import jsonify


class Error(Exception):
    """Base error class."""


class ApiHTTPError(Error):
    """Error class for API HTTP errors."""

    def __init__(self, message, status_code):
        """Initialize the class with a custom error message.

        Args:
            message: Description of the error.
            status_code: HTTP status code.
        """
        super(ApiHTTPError, self).__init__()
        self.message = message
        self.status_code = status_code

    def build_response(self):
        """Create a response object.

        Returns:
            Response object (instance of flask.wrappers.Response)
        """
        response = jsonify({
            'message': self.message,
            'status': self.status_code
        })
        response.status_code = self.status_code
        return response


class DataIngestionError(Error):
    """Raised when unable to ingest data."""



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements HTTP request handler."""

from __future__ import unicode_literals

from flask import Blueprint
from flask import redirect
from flask import render_template
from flask_login import login_required

# Register flask blueprint
spa_views = Blueprint('spa_views', __name__)


@spa_views.route(
    '/sketch/<int:sketch_id>/explore/view/<int:view_id>/', methods=['GET'])
@login_required
# pylint: disable=unused-argument
def redirect_view(sketch_id, view_id):
    """Redirect old (deprecated) view URLs to scheme.

    Returns:
        Redirect to new URL scheme.
    """
    return redirect(
        '/sketch/{0:d}/explore?view={1:d}'.format(sketch_id, view_id))


@spa_views.route('/', defaults={'path': ''})
@spa_views.route('/<path:path>')
@login_required
# pylint: disable=unused-argument
def overview(path):
    """Generates the template.

    Returns:
        Template with context.
    """
    return render_template('index.html')



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing end to end tests."""


class EndToEndTestManager(object):
    """The test manager."""

    _class_registry = {}
    _exclude_registry = set()

    @classmethod
    def get_tests(cls):
        """Retrieves the registered tests.

        Yields:
            tuple: containing:
                unicode: the uniquely identifying name of the test
                type: the test class.
        """
        for test_name, test_class in iter(cls._class_registry.items()):
            if test_name in cls._exclude_registry:
                continue
            yield test_name, test_class

    @classmethod
    def get_test(cls, test_name):
        """Retrieves a class object of a specific test.

        Args:
            test_name (unicode): name of the test to retrieve.

        Returns:
            Instance of Test class object.

        Raises:
            KeyError: if the test is not registered.
        """
        try:
            test_class = cls._class_registry[test_name.lower()]
        except KeyError:
            raise KeyError(
                'No such test type: {0:s}'.format(test_name.lower()))
        return test_class

    @classmethod
    def register_test(cls, test_class, exclude_from_list=False):
        """Registers an test class.

        The test classes are identified by their lower case name.

        Args:
            test_class (type): the test class to register.
            exclude_from_list (boolean): if set to True then the test
                gets registered but will not be included in the
                get_tests function. Defaults to False.

        Raises:
            KeyError: if class is already set for the corresponding name.
        """
        test_name = test_class.NAME.lower()
        if test_name in cls._class_registry:
            raise KeyError('Class already set for name: {0:s}.'.format(
                test_class.NAME))
        cls._class_registry[test_name] = test_class
        if exclude_from_list:
            cls._exclude_registry.add(test_name)

    @classmethod
    def clear_registration(cls):
        """Clears all test registrations."""
        cls._class_registry = {}
        cls._exclude_registry = set()



"""Psexec sequence activity sessionizing sketch analyzer plugins."""

from __future__ import unicode_literals

from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import sequence_sessionizer

# TODO Add PSEXEC sequence sessionizer for the source side.

DEST_SEQ = [{
    'data_type': 'windows:evtx:record',
    'message': 'A service was installed in the system.'
}, {
    'data_type': 'windows:prefetch:execution',
    'executable': 'PSEXESVC.EXE'
}]


class DestPsexecSessionizerSketchPlugin(
        sequence_sessionizer.SequenceSessionizerSketchPlugin):
    """Sessionizer for seqeunced psexec activity on the server."""
    NAME = 'dest_psexec_sessionizer'
    query = 'PSEXEC OR PSEXESVC OR PSEXESVC.EXE'
    event_seq = DEST_SEQ
    session_type = 'psexec_dest'
    return_fields = None


manager.AnalysisManager.register_analyzer(DestPsexecSessionizerSketchPlugin)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functions and utilities."""

from __future__ import unicode_literals

import colorsys
import csv
import datetime
import email
import json
import logging
import random
import smtplib
import sys
import time
import codecs
import six

from dateutil import parser
from flask import current_app

from timesketch.lib import errors

logger = logging.getLogger('timesketch.utils')

# Set CSV field size limit to systems max value.
csv.field_size_limit(sys.maxsize)

# Fields to scrub from timelines.
FIELDS_TO_REMOVE = ['_id', '_type', '_index', '_source']


def random_color():
    """Generates a random color.

    Returns:
        Color as string in HEX
    """
    hue = random.random()
    golden_ratio_conjugate = (1 + 5**0.5) / 2
    hue += golden_ratio_conjugate
    hue %= 1
    rgb = tuple(int(i * 256) for i in colorsys.hsv_to_rgb(hue, 0.5, 0.95))
    return '{0:02X}{1:02X}{2:02X}'.format(rgb[0], rgb[1], rgb[2])


def _parse_tag_field(row):
    """Reading in a tag field and converting to a list of strings."""
    if isinstance(row, (list, tuple)):
        return row

    if not isinstance(row, str):
        row = str(row)

    if row.startswith('[') and row.endswith(']'):
        return json.loads(row)

    if row == '-':
        return []

    if ',' in row:
        return row.split(',')

    return [row]


def _scrub_special_tags(dict_obj):
    """Remove Elastic specific fields from a dict."""
    for field in FIELDS_TO_REMOVE:
        if field in dict_obj:
            _ = dict_obj.pop(field)


def read_and_validate_csv(file_handle, delimiter=','):
    """Generator for reading a CSV file.

    Args:
        file_handle: a file-like object containing the CSV content.
        delimiter: character used as a field separator, default: ','

    Raises:
        RuntimeError: when there are missing fields.
        DataIngestionError: when there are issues with the data ingestion.
    """
    # Columns that must be present in the CSV file.
    mandatory_fields = ['message', 'datetime', 'timestamp_desc']

    # Ensures delimiter is a string.
    if not isinstance(delimiter, six.text_type):
        delimiter = codecs.decode(delimiter, 'utf8')

    # Due to issues with python2.
    if six.PY2:
        delimiter = str(delimiter)

    reader = csv.DictReader(file_handle, delimiter=delimiter)
    csv_header = reader.fieldnames
    missing_fields = []
    # Validate the CSV header
    for field in mandatory_fields:
        if field not in csv_header:
            missing_fields.append(field)
    if missing_fields:
        raise RuntimeError(
            'Missing fields in CSV header: {0:s}'.format(
                ','.join(missing_fields)))
    try:
        for row in reader:
            # There is a condition in which the CSV reader can read a
            # single lines as multiple, causing issues with importing.
            # TODO: Swap the CSV library for the user of pandas.
            if not row:
                continue
            if not row['datetime']:
                logger.warning(
                    'Row missing a datetime object, skipping [{0:s}]'.format(
                        ','.join([str(x).replace(
                            '\n', '').strip() for x in row.values()])))
                continue
            try:
                # normalize datetime to ISO 8601 format if it's not the case.
                parsed_datetime = parser.parse(row['datetime'])
                row['datetime'] = parsed_datetime.isoformat()

                normalized_timestamp = int(
                    time.mktime(parsed_datetime.utctimetuple()) * 1000000)
                normalized_timestamp += parsed_datetime.microsecond
                row['timestamp'] = str(normalized_timestamp)
                if 'tag' in row:
                    row['tag'] = [x for x in _parse_tag_field(row['tag']) if x]

                _scrub_special_tags(row)
            except ValueError:
                continue

            yield row
    except csv.Error as e:
        error_string = 'Unable to read file, with error: {0!s}'.format(e)
        logger.error(error_string)
        raise errors.DataIngestionError(error_string)


def read_and_validate_redline(file_handle):
    """Generator for reading a Redline CSV file.
    Args:
        file_handle: a file-like object containing the CSV content.
    """
    # Columns that must be present in the CSV file
    mandatory_fields = ['Alert', 'Tag', 'Timestamp', 'Field', 'Summary']

    csv.register_dialect(
        'myDialect', delimiter=',', quoting=csv.QUOTE_ALL,
        skipinitialspace=True)
    reader = csv.DictReader(file_handle, delimiter=',', dialect='myDialect')

    csv_header = reader.fieldnames
    missing_fields = []
    # Validate the CSV header
    for field in mandatory_fields:
        if field not in csv_header:
            missing_fields.append(field)
    if missing_fields:
        raise RuntimeError(
            'Missing fields in CSV header: {0:s}'.format(missing_fields))
    for row in reader:
        dt = parser.parse(row['Timestamp'])
        timestamp = int(time.mktime(dt.timetuple())) * 1000
        dt_iso_format = dt.isoformat()
        timestamp_desc = row['Field']

        summary = row['Summary']
        alert = row['Alert']
        tag = row['Tag']

        row_to_yield = {}
        row_to_yield["message"] = summary
        row_to_yield["timestamp"] = timestamp
        row_to_yield["datetime"] = dt_iso_format
        row_to_yield["timestamp_desc"] = timestamp_desc
        row_to_yield["alert"] = alert #extra field
        tags = [tag]
        row_to_yield["tag"] = tags # extra field

        yield row_to_yield


def read_and_validate_jsonl(file_handle):
    """Generator for reading a JSONL (json lines) file.

    Args:
        file_handle: a file-like object containing the CSV content.

    Raises:
        RuntimeError: if there are missing fields.
        DataIngestionError: If the ingestion fails.

    Yields:
        A dict that's ready to add to the datastore.
    """
    # Fields that must be present in each entry of the JSONL file.
    mandatory_fields = ['message', 'datetime', 'timestamp_desc']
    lineno = 0
    for line in file_handle:
        lineno += 1
        try:
            linedict = json.loads(line)
            ld_keys = linedict.keys()
            if 'datetime' not in ld_keys and 'timestamp' in ld_keys:
                epoch = int(str(linedict['timestamp'])[:10])
                dt = datetime.datetime.fromtimestamp(epoch)
                linedict['datetime'] = dt.isoformat()
            if 'timestamp' not in ld_keys and 'datetime' in ld_keys:
                linedict['timestamp'] = parser.parse(linedict['datetime'])

            missing_fields = [x for x in mandatory_fields if x not in linedict]
            if missing_fields:
                raise RuntimeError(
                    'Missing field(s) at line {0:n}: {1:s}'.format(
                        lineno, ','.join(missing_fields)))

            if 'tag' in linedict:
                linedict['tag'] = [
                    x for x in _parse_tag_field(linedict['tag']) if x]
            _scrub_special_tags(linedict)
            yield linedict

        except ValueError as e:
            raise errors.DataIngestionError(
                'Error parsing JSON at line {0:n}: {1:s}'.format(lineno, e))


def get_validated_indices(indices, sketch_indices):
    """Exclude any deleted search index references.

    Args:
        indices: List of indices from the user
        sketch_indices: List of indices in the sketch

    Returns:
        Set of indices with those removed that is not in the sketch
    """
    exclude = set(indices) - set(sketch_indices)
    if exclude:
        indices = [index for index in indices if index not in exclude]

    return indices


def send_email(subject, body, to_username, use_html=False):
    """Send email using configure SMTP server.

    Args:
        subject: Email subject string.
        body: Email message body.
        to_username: User to send email to.
        use_html: Boolean indicating if the email body should be sent as html.

    Raises:
        RuntimeError if not properly configured or if the recipient user is no
        in the whitelist.
    """
    email_enabled = current_app.config.get('ENABLE_EMAIL_NOTIFICATIONS')
    email_domain = current_app.config.get('EMAIL_DOMAIN')
    email_smtp_server = current_app.config.get('EMAIL_SMTP_SERVER')
    email_from_user = current_app.config.get('EMAIL_FROM_ADDRESS', 'timesketch')
    email_user_whitelist = current_app.config.get('EMAIL_USER_WHITELIST', [])

    if not email_enabled:
        raise RuntimeError('Email notifications are not enabled, aborting.')

    if not email_domain:
        raise RuntimeError('Email domain is not configured, aborting.')

    if not email_smtp_server:
        raise RuntimeError('Email SMTP server is not configured, aborting.')

    # Only send mail to whitelisted usernames.
    if to_username not in email_user_whitelist:
        return

    from_address = '{0:s}@{1:s}'.format(email_from_user, email_domain)
    # TODO: Add email address to user object and pick it up from there.
    to_address = '{0:s}@{1:s}'.format(to_username, email_domain)
    email_content_type = 'text'
    if use_html:
        email_content_type = 'text/html'

    msg = email.message.Message()
    msg['Subject'] = subject
    msg['From'] = from_address
    msg['To'] = to_address
    msg.add_header('Content-Type', email_content_type)
    msg.set_payload(body)

    smtp = smtplib.SMTP(email_smtp_server)
    smtp.sendmail(msg['From'], [msg['To']], msg.as_string())
    smtp.quit()



"""Tests for ThreatintelPlugin."""
from __future__ import unicode_literals

import copy
import re
import mock

from flask import current_app

from timesketch.lib.analyzers import yetiindicators
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

MOCK_YETI_INTEL = {
    'x-regex--6ebc9344-1111-4d65-8bdd-b6dddf613068': {
        'id': 'x-regex--6ebc9344-1111-4d65-8bdd-b6dddf613068',
        'name': 'Secret Fancy Bear c2',
        'pattern': 'baddomain\\.com',
        'compiled_regexp': re.compile('baddomain\\.com'),
        'type': 'x-regex',
    }
}

MOCK_YETI_NEIGHBORS = [{
    'id': 'x-incident--6ebc9344-1111-4d65-8bdd-b6dddf613068',
    'name': 'Random incident',
    'type': 'x-incident',
}]

MATCHING_DOMAIN_MESSAGE = {'message': 'baddomain.com'}
OK_DOMAIN_MESSAGE = {'message': 'okdomain.com'}



class TestThreatintelPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def setUp(self):
        super(TestThreatintelPlugin, self).setUp()
        current_app.config['YETI_API_ROOT'] = 'blah'
        current_app.config['YETI_API_KEY'] = 'blah'

    # Mock the Elasticsearch datastore.
    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    @mock.patch('timesketch.lib.analyzers.yetiindicators.'
                'YetiIndicators.get_neighbors')
    @mock.patch('timesketch.lib.analyzers.yetiindicators.'
                'YetiIndicators.get_indicators')
    def test_indicator_match(self, mock_get_indicators, mock_get_neighbors):
        """Test that ES queries for indicators are correctly built."""
        sessionizer = yetiindicators.YetiIndicators('test_index', 1)
        sessionizer.datastore.client = mock.Mock()
        sessionizer.intel = MOCK_YETI_INTEL
        mock_get_neighbors.return_value = MOCK_YETI_NEIGHBORS

        event = copy.deepcopy(MockDataStore.event_dict)
        event['_source'].update(MATCHING_DOMAIN_MESSAGE)
        sessionizer.datastore.import_event(
            'test_index', event['_type'], event['_source'], '0')

        message = sessionizer.run()
        self.assertEqual(
            message,
            '1 events matched 1 indicators. [Random incident:x-incident]')
        mock_get_indicators.assert_called_once()
        mock_get_neighbors.assert_called_once()

    # Mock the Elasticsearch datastore.
    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    @mock.patch('timesketch.lib.analyzers.yetiindicators.'
                'YetiIndicators.get_neighbors')
    @mock.patch('timesketch.lib.analyzers.yetiindicators.'
                'YetiIndicators.get_indicators')
    def test_indicator_nomatch(self, mock_get_indicators, mock_get_neighbors):
        """Test that ES queries for indicators are correctly built."""
        sessionizer = yetiindicators.YetiIndicators('test_index', 1)
        sessionizer.datastore.client = mock.Mock()
        sessionizer.intel = MOCK_YETI_INTEL
        mock_get_neighbors.return_value = MOCK_YETI_NEIGHBORS

        event = copy.deepcopy(MockDataStore.event_dict)
        event['_source'].update(OK_DOMAIN_MESSAGE)
        sessionizer.datastore.import_event(
            'test_index', event['_type'], event['_source'], '0')

        message = sessionizer.run()
        self.assertEqual(message, 'No indicators were found in the timeline.')
        mock_get_indicators.assert_called_once()
        mock_get_neighbors.assert_not_called()

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_slug(self):
        sessionizer = yetiindicators.YetiIndicators('test_index', 1)
        mock_event = mock.Mock()
        sessionizer.mark_event(
            MOCK_YETI_INTEL['x-regex--6ebc9344-1111-4d65-8bdd-b6dddf613068'],
            mock_event,
            MOCK_YETI_NEIGHBORS)
        # The name of the entity is "Random incident"
        mock_event.add_tags.assert_called_once_with(['random-incident'])



"""Sketch analyzer plugin for ntfs timestomping detection."""
from __future__ import unicode_literals

from flask import current_app

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class FileInfo(object):
    """Datastructure to track all timestamps for a file and timestamp type."""
    def __init__(self, file_reference=None, timestamp_desc=None,
                 std_info_event=None, std_info_timestamp=None, file_names=None):
        self.file_reference = file_reference
        self.timestamp_desc = timestamp_desc
        self.std_info_event = std_info_event
        self.std_info_timestamp = std_info_timestamp
        self.file_names = file_names or []


class NtfsTimestompSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for Timestomp."""

    NAME = 'ntfs_timestomp'
    STD_INFO = 16
    FILE_NAME = 48

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        self.threshold = current_app.config.get(
            'NTFS_TIMESTOMP_ANALYZER_THRESHOLD', 10) * 60000000
        super(NtfsTimestompSketchPlugin, self).__init__(index_name, sketch_id)

    def is_suspicious(self, file_info):
        """Compares timestamps to detect timestomping.

        Args:
            file_info: FileInfo object for event to look at.

        Returns:
            Boolean, true if timestomping was detected.

        """
        if not file_info.std_info_event or not file_info.file_names:
            return False

        suspicious = True
        diffs = []

        for fn in file_info.file_names:
            diff = fn[1] - file_info.std_info_timestamp
            diffs.append(diff)

            if abs(diff) > self.threshold:
                fn[0].add_attributes({'time_delta': diff})
            else:
                suspicious = False
                break

        if suspicious:
            for fn in file_info.file_names:
                fn[0].commit()

            file_info.std_info_event.add_attributes({'time_deltas': diffs})
            file_info.std_info_event.commit()

        return suspicious

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result.
        """

        query = 'attribute_type:48 OR attribute_type:16'

        return_fields = ['attribute_type', 'timestamp_desc',
                         'file_reference', 'timestamp']

        events = self.event_stream(
            query_string=query, return_fields=return_fields)

        # Dict timestamp_type + "&" + file_ref -> FileInfo
        file_infos = dict()

        for event in events:
            attribute_type = event.source.get('attribute_type')
            file_ref = event.source.get('file_reference')
            timestamp_type = event.source.get('timestamp_desc')
            timestamp = event.source.get('timestamp')

            if not attribute_type or not timestamp_type:
                continue

            if attribute_type not in [self.FILE_NAME, self.STD_INFO]:
                continue

            key = '{0:s}&{1:s}'.format(timestamp_type, str(file_ref))

            if key not in file_infos:
                file_infos[key] = FileInfo()

            file_info = file_infos[key]
            file_info.file_reference = file_ref
            file_info.timestamp_desc = timestamp_type

            if attribute_type == self.STD_INFO:
                file_info.std_info_timestamp = timestamp
                file_info.std_info_event = event

            if attribute_type == self.FILE_NAME:
                file_info.file_names.append((event, timestamp))

        timestomps = 0
        for file_info in file_infos.values():
            if self.is_suspicious(file_info):
                timestomps = timestomps + 1

        if timestomps > 0:
            self.sketch.add_view(
                view_name='NtfsTimestomp', analyzer_name=self.NAME,
                query_string='_exists_:time_delta or _exists:time_deltas')

        return ('NtfsTimestomp Analyzer done, found {0:d} timestomped events'
                .format(timestomps))


manager.AnalysisManager.register_analyzer(NtfsTimestompSketchPlugin)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Session resources for version 1 of the Timesketch API."""

from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models.sketch import Sketch


class SessionResource(resources.ResourceMixin, Resource):
    """Resource to get sessions."""

    @login_required
    def get(self, sketch_id, timeline_index):
        """Handles GET request to the resource.

        Returns:
            A list of objects representing sessions.
        """
        max_ids = 10000  # More than the number of sessions we expect to return
        max_sessions = 100

        session_types = ['all_events', 'web_activity', 'logon_session',
                         'ssh_bruteforce_session', 'ssh_session']
        sessions = []
        is_truncated = False

        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        # Check the timeline belongs to the sketch
        sketch_indices = {t.searchindex.index_name for t in sketch.timelines
                          if t.searchindex.index_name == timeline_index}

        id_agg_spec = {
            'aggregations': {
                'term_count': {
                    'terms': {
                        'field': '',
                        'size': max_ids
                    }
                }
            }
        }

        ts_agg_spec = {
            'aggregations': {
                'timestamp_range': {
                    'filter': {
                        'bool': {
                            'must': [{
                                'query_string': {
                                    'query': ''
                                }
                            }]
                        }
                    },
                    'aggregations': {
                        'min_timestamp': {
                            'min': {
                                'field': 'timestamp'
                            }
                        },
                        'max_timestamp': {
                            'max': {
                                'field': 'timestamp'
                            }
                        }
                    }
                }
            }
        }

        id_terms = id_agg_spec['aggregations']['term_count']['terms']
        ts_filter = ts_agg_spec['aggregations']['timestamp_range']['filter']
        ts_query_string = ts_filter['bool']['must'][0]['query_string']

        for session_type in session_types:
            id_terms['field'] = 'session_id.{}.keyword'.format(session_type)
            # pylint: disable=unexpected-keyword-arg
            id_agg = self.datastore.client.search(index=list(sketch_indices),
                                                  body=id_agg_spec,
                                                  size=0)
            buckets = id_agg['aggregations']['term_count']['buckets']
            session_count = 0

            for bucket in buckets:
                if session_count == max_sessions:
                    is_truncated = True
                    break
                session_count += 1

                session_id = bucket['key']
                ts_query_string['query'] = 'session_id.{}:{}'.format(
                    session_type,
                    session_id)
                ts_agg = self.datastore.client.search(
                    index=list(sketch_indices),
                    body=ts_agg_spec,
                    size=0)
                start_timestamp = int(ts_agg['aggregations']
                                      ['timestamp_range']['min_timestamp']
                                      ['value']) / 1000
                end_timestamp = int(ts_agg['aggregations']
                                    ['timestamp_range']['max_timestamp']
                                    ['value']) / 1000

                sessions.append({'session_type': session_type,
                                 'session_id': session_id,
                                 'start_timestamp': start_timestamp,
                                 'end_timestamp': end_timestamp})

        sessions.append({'truncated': is_truncated})
        return sessions



"""The sketch analyzer for chained events."""
from __future__ import unicode_literals

import collections
import uuid

from timesketch.lib import emojis
from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import chain_plugins  # pylint: disable=unused-import
from timesketch.lib.analyzers.chain_plugins import manager as chain_manager


class ChainSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for chained events.

    The purpose of the chain analyzer is to chain together events that can
    be described as linked, either by sharing some common entitites, or
    one event being a derivative of another event. An example of this
    would be that a browser downloads an executable, which then later gets
    executed. The signs of execution could lie in multiple events, from
    different sources, but they are all linked or chained together. This
    could help an analyst see the connection between these separate but
    chained events. Another example could be a document written and then
    compressed into a ZIP file, which would then be exfilled through some
    means. If the document and the ZIP file are chained together it could be
    easier for the analyst to track the meaning of an exfil event involving the
    compressed file.
    """

    NAME = 'chain'

    DEPENDENCIES = frozenset()

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        self._chain_plugins = (
            chain_manager.ChainPluginsManager.get_plugins(self))
        super(ChainSketchPlugin, self).__init__(index_name, sketch_id)


    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        link_emoji = emojis.get_emoji('LINK')

        number_of_base_events = 0
        number_of_chains = 0
        counter = collections.Counter()
        events_to_update = {}

        # TODO: Have each plugin run in a separate task.
        # TODO: Add a time limit for each plugins run to prevent it from
        #       holding everything up.
        for chain_plugin in self._chain_plugins:
            if chain_plugin.SEARCH_QUERY_DSL:
                search_dsl = chain_plugin.SEARCH_QUERY_DSL
                search_string = None
            else:
                search_dsl = None
                search_string = chain_plugin.SEARCH_QUERY

            return_fields = chain_plugin.EVENT_FIELDS
            events = self.event_stream(
                query_string=search_string, query_dsl=search_dsl,
                return_fields=return_fields)

            for event in events:
                if not chain_plugin.process_chain(event):
                    continue
                chain_id = uuid.uuid4().hex

                chained_events = chain_plugin.build_chain(
                    base_event=event, chain_id=chain_id)
                number_chained_events = len(chained_events)
                if not number_chained_events:
                    continue

                for chained_event in chained_events:
                    chained_id = chained_event.get('event_id')
                    if chained_id not in events_to_update:
                        default = {
                            'event': chained_event.get('event'),
                            'chains': []
                        }
                        events_to_update[chained_id] = default

                    events_to_update[chained_id]['chains'].append(
                        chained_event.get('chain'))

                number_of_base_events += 1

                counter[chain_plugin.NAME] += number_chained_events
                counter['total'] += number_chained_events

                chain = {
                    'chain_id': chain_id,
                    'plugin': chain_plugin.NAME,
                    'is_base': True,
                    'leafs': number_chained_events,
                }
                if event.event_id not in events_to_update:
                    default = {
                        'event': event,
                        'chains': []
                    }
                    events_to_update[event.event_id] = default
                events_to_update[event.event_id]['chains'].append(chain)
                number_of_chains += 1

        for event_update in events_to_update.values():
            event = event_update.get('event')
            attributes = {
                'chains': event_update.get('chains')
            }
            event.add_attributes(attributes)
            event.add_emojis([link_emoji])
            event.commit()

        chain_string = ' - '.join([
            '[{0:s}] {1:d}'.format(
                x[0], x[1]) for x in counter.most_common() if x[0] != 'total'])
        return (
            '{0:d} base events annotated with a chain UUID for {1:d} '
            'chains for a total of {2:d} events. {3:s}'.format(
                number_of_base_events, number_of_chains,
                counter['total'], chain_string))


manager.AnalysisManager.register_analyzer(ChainSketchPlugin)



"""Tests for DomainsPlugin."""
from __future__ import unicode_literals

import mock

from flask import current_app
from datasketch.minhash import MinHash

from timesketch.lib.analyzers import phishy_domains
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestDomainsPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def setUp(self):
        """Set up the tests."""
        super(TestDomainsPlugin, self).setUp()
        current_app.config['DOMAIN_ANALYZER_WATCHED_DOMAINS'] = ['foobar.com']
        current_app.config['DOMAIN_ANALYZER_WATCHED_DOMAINS_THRESHOLD'] = 10
        current_app.config[
            'DOMAIN_ANALYZER_WATCHED_DOMAINS_SCORE_THRESHOLD'] = 0.75
        current_app.config[
            'DOMAIN_ANALYZER_WHITELISTED_DOMAINS'] = [
                'ytimg.com', 'gstatic.com', 'yimg.com',
                'akamaized.net', 'akamaihd.net', 's-microsoft.com']

    # Mock the Elasticsearch datastore.
    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_minhash(self):
        """Test minhash function."""
        analyzer = phishy_domains.PhishyDomainsSketchPlugin('test_index', 1)
        domain = 'www.mbl.is'
        # pylint: disable=protected-access
        minhash = analyzer._get_minhash_from_domain(domain)

        self.assertIsInstance(minhash, MinHash)

        another_domain = 'mbl.is'
        # pylint: disable=protected-access
        minhash2 = analyzer._get_minhash_from_domain(another_domain)

        self.assertEqual(minhash.jaccard(minhash2), 0.546875)

    # Mock the Elasticsearch datastore.
    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_get_similar_domains(self):
        """Test get_similar_domains function."""
        analyzer = phishy_domains.PhishyDomainsSketchPlugin('test_index', 1)
        domain = 'login.stortmbl.is'
        # pylint: disable=protected-access
        minhash = analyzer._get_minhash_from_domain(domain)
        domain_dict = {domain: {'hash': minhash, 'depth': 3}}

        # pylint: disable=protected-access
        similar = analyzer._get_similar_domains(
            'login.stortmbi.is', domain_dict)
        self.assertEqual(len(similar), 1)

        # pylint: disable=protected-access
        similar = analyzer._get_similar_domains('www.google.com', domain_dict)
        self.assertEqual(len(similar), 0)



#!/usr/bin/env python
# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This is the setup file for the project. The standard setup rules apply:

   python setup.py build
   sudo python setup.py install
"""

from __future__ import print_function
from __future__ import unicode_literals

import glob
import os
import sys

import pkg_resources

from setuptools import find_packages
from setuptools import setup

from timesketch import version


version_tuple = (sys.version_info[0], sys.version_info[1])
if version_tuple < (3, 6):
    print((
        'Unsupported Python version: {0:s}, version 3.6 or higher '
        'required.').format(sys.version))
    sys.exit(1)


def parse_requirements_from_file(path):
    """Parses requirements from a requirements file.

    Args:
      path (str): path to the requirements file.

    Yields:
      pkg_resources.Requirement: package resource requirement.
    """
    with open(path, 'r') as file_object:
        file_contents = file_object.read()
    for req in pkg_resources.parse_requirements(file_contents):
        try:
            requirement = str(req.req)
        except AttributeError:
            requirement = str(req)
        yield requirement


timesketch_description = (
    'Timesketch is a web based tool for collaborative forensic timeline '
    'analysis. Using sketches you and your collaborators can easily organize '
    'timelines and analyze them all at the same time.  Add meaning to '
    'your raw data with rich annotations, comments, tags and stars.')

setup(
    name='timesketch',
    version=version.get_version(),
    description='Digital forensic timeline analysis',
    long_description=timesketch_description,
    license='Apache License, Version 2.0',
    url='http://www.timesketch.org/',
    maintainer='Timesketch development team',
    maintainer_email='timesketch-dev@googlegroups.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    data_files=[
        ('share/timesketch', glob.glob(
            os.path.join('data', '*.*'))),
        ('share/timesketch/linux', glob.glob(
            os.path.join('data', 'linux', '*.*'))),
        ('share/doc/timesketch', [
            'AUTHORS', 'LICENSE', 'README.md']),
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={'console_scripts': ['tsctl=timesketch.tsctl:main']},
    install_requires=parse_requirements_from_file('requirements.txt'),
    tests_require=parse_requirements_from_file('test_requirements.txt'),
)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module implements annotations that can be use on other database models.
"""

from __future__ import unicode_literals

import json
import six

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Unicode
from sqlalchemy import UnicodeText
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

from timesketch.models import BaseModel
from timesketch.models import db_session


class BaseAnnotation(object):
    """Base class with common attributes."""

    @declared_attr
    def user_id(self):
        """Foreign key to a user model.

        Returns:
            A column (instance of sqlalchemy.Column)
        """
        return Column(Integer, ForeignKey('user.id'))

    @declared_attr
    def user(self):
        """A relationship to a user object.

        Returns:
            A relationship (instance of sqlalchemy.orm.relationship)
        """
        return relationship('User')


class Label(BaseAnnotation):
    """A label annotation."""
    label = Column(Unicode(255))

    def __init__(self, user, label):
        """Initialize the model.

        Args:
            user: A user (instance of timesketch.models.user.User)
            name: Name of the label
        """
        super(Label, self).__init__()
        self.user = user
        self.label = label


class Comment(BaseAnnotation):
    """A comment annotation."""
    comment = Column(UnicodeText())

    def __init__(self, user, comment):
        """Initialize the model.

        Args:
            user: A user (instance of timesketch.models.user.User)
            body: The body if the comment
        """
        super(Comment, self).__init__()
        self.user = user
        self.comment = comment


class Status(BaseAnnotation):
    """A status annotation."""
    status = Column(Unicode(255))

    def __init__(self, user, status):
        """Initialize the model.

        Args:
            user: A user (instance of timesketch.models.user.User)
            status: The type of status (string, e.g. open)
        """
        super(Status, self).__init__()
        self.user = user
        self.status = status


class LabelMixin(object):
    """
    A MixIn for generating the necessary tables in the database and to make
    it accessible from the parent model object (the model object that uses this
    MixIn, i.e. the object that the label is added to).
    """

    @declared_attr
    def labels(self):
        """
        Generates the label tables and adds the attribute to the parent model
        object.

        Returns:
            A relationship to an label (timesketch.models.annotation.Label)
        """
        if six.PY2:
            class_name = b'{0:s}Label'.format(self.__name__)
        else:
            class_name = '{0:s}Label'.format(self.__name__)

        self.Label = type(class_name, (
            Label,
            BaseModel,),
                          dict(
                              __tablename__='{0:s}_label'.format(
                                  self.__tablename__),
                              parent_id=Column(
                                  Integer,
                                  ForeignKey('{0:s}.id'.format(
                                      self.__tablename__))),
                              parent=relationship(self)))
        return relationship(self.Label)

    def add_label(self, label, user=None):
        """Add a label to an object.

        Each entry can have multible labels.

        Args:
            label: Name of the label.
            user: Optional user that adds the label (sketch.User).
        """
        if self.has_label(label):
            return
        self.labels.append(self.Label(user=user, label=label))
        db_session.commit()

    def remove_label(self, label):
        """Remove a label from an object.

        Args:
            label: Name of the label.
        """
        for label_obj in self.labels:
            if label_obj.label.lower() != label.lower():
                continue
            self.labels.remove(label_obj)
        db_session.commit()

    def has_label(self, label):
        """Returns a boolean whether a label is applied.

        Args:
            label: Name of the label.

        Returns:
            True if the label is set, False otherwise.
        """
        for label_obj in self.labels:
            if label_obj.label.lower() == label.lower():
                return True

        return False

    @property
    def get_labels(self):
        """Returns a list of all applied labels.

        Returns:
            A list of strings with all the applied labels.
        """
        if not self.labels:
            return []

        return [x.label for x in self.labels]

    @property
    def label_string(self):
        """Returns a JSON encoded string with a list of the labels.

        Returns:
            A JSON encoded string with the list of labels.
        """
        if not self.labels:
            return ''

        return json.dumps([x.label for x in self.labels])


class CommentMixin(object):
    """
    A MixIn for generating the necessary tables in the database and to make
    it accessible from the parent model object (the model object that uses this
    MixIn, i.e. the object that the comment is added to).
    """

    @declared_attr
    def comments(self):
        """
        Generates the comment tables and adds the attribute to the parent model
        object.

        Returns:
            A relationship to a comment (timesketch.models.annotation.Comment)
        """
        if six.PY2:
            class_name = b'{0:s}Comment'.format(self.__name__)
        else:
            class_name = '{0:s}Comment'.format(self.__name__)

        self.Comment = type(
            class_name, (
                Comment,
                BaseModel, ),
            dict(
                __tablename__='{0:s}_comment'.format(self.__tablename__),
                parent_id=Column(
                    Integer,
                    ForeignKey('{0:s}.id'.format(self.__tablename__))),
                parent=relationship(self), ))
        return relationship(self.Comment)


class StatusMixin(object):
    """
    A MixIn for generating the necessary tables in the database and to make
    it accessible from the parent model object (the model object that uses this
    MixIn, i.e. the object that the status is added to).
    """

    @declared_attr
    def status(self):
        """
        Generates the status tables and adds the attribute to the parent model
        object.

        Returns:
            A relationship to a status (timesketch.models.annotation.Status)
        """
        if six.PY2:
            class_name = b'{0:s}Status'.format(self.__name__)
        else:
            class_name = '{0:s}Status'.format(self.__name__)

        self.Status = type(
            class_name, (
                Status,
                BaseModel, ),
            dict(
                __tablename__='{0:s}_status'.format(self.__tablename__),
                parent_id=Column(
                    Integer,
                    ForeignKey('{0:s}.id'.format(self.__tablename__))),
                parent=relationship(self), ))
        return relationship(self.Status)

    def set_status(self, status):
        """
        Set status on object. Although this is a many-to-many relationship
        this makes sure that the parent object only has one status set.

        Args:
            status: Name of the status
        """
        for _status in self.status:
            self.status.remove(_status)
        self.status.append(self.Status(user=None, status=status))
        db_session.commit()

    @property
    def get_status(self):
        """Get the current status.

        Returns:
            The status as a string
        """
        if not self.status:
            self.status.append(self.Status(user=None, status='new'))
        return self.status[0]



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ontology for attributes as well as related functions."""

from timesketch.lib.analyzers import interface


def ontology():
    """Return a dict with the ontology definitions."""
    return interface.get_yaml_config('ontology.yaml')


def cast_variable(value, cast_as_str):
    """Cast a variable and return it.

    Args:
      value (str): Value as a string.
      cast_as_str (str): The type to cast it as.

    Raises
      TypeError: If the value isn't already a string.

    Returns:
      The value cast as cast_as_str defines.
    """
    if not isinstance(value, str):
        raise TypeError('Value needs to be a string, not {0!s}'.format(
            type(value)))

    if cast_as_str == 'str':
        return value

    if cast_as_str == 'int':
        return int(value)

    if cast_as_str == 'float':
        return float(value)

    if cast_as_str == 'bool':
        return bool(value == 'True')

    # TODO: Support more casting.
    return value


ONTOLOGY = ontology()



"""Tests for SafeBrowsingSketchPlugin."""
from __future__ import unicode_literals

from collections import OrderedDict

import mock
from httmock import HTTMock, all_requests

from timesketch.lib.analyzers import safebrowsing
from timesketch.lib.testlib import BaseTest, MockDataStore


class TestSafeBrowsingSketchPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore,
    )
    def test_safebrowsing_analyzer_class(self):
        """Test core functionality of the analyzer class."""
        index_name = 'test'
        sketch_id = 1
        analyzer = safebrowsing.SafeBrowsingSketchPlugin(index_name, sketch_id)
        self.assertEqual(analyzer.index_name, index_name)
        self.assertEqual(analyzer.sketch.id, sketch_id)

    @all_requests
    # pylint: disable=unused-argument,missing-docstring
    def safebrowsing_find_mock(self, url, request):
        MOCK_RESULT = {
            'matches': [
                {
                    'threat': {
                        'url': 'http://A',
                    },
                    'cacheDuration': '300s',
                    'threatEntryType': 'URL',
                    'threatType': 'MALWARE',
                    'platformType': 'ANY_PLATFORM',
                },
                {
                    'threat': {
                        'url': 'https://B',
                    },
                    'cacheDuration': '300s',
                    'threatEntryType': 'URL',
                    'threatType': 'MALWARE',
                    'platformType': 'WINDOWS',
                },
            ],
        }

        return {
            'status_code': 200,
            'content': MOCK_RESULT,
        }

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore,
    )
    # pylint: disable=missing-docstring
    def test_do_safebrowsing_lookup(self):
        index_name = 'test'
        sketch_id = 1
        analyzer = safebrowsing.SafeBrowsingSketchPlugin(index_name, sketch_id)

        with HTTMock(self.safebrowsing_find_mock):
            EXPECTED_RESULT = {
                'http://A': {
                    'platformType': 'ANY_PLATFORM',
                    'threatType': 'MALWARE',
                },
                'https://B': {
                    'platformType': 'WINDOWS',
                    'threatType': 'MALWARE',
                },
            }

            # pylint: disable=protected-access
            actual_result = analyzer._do_safebrowsing_lookup(
                ['http://A', 'https://B'],
                [],
                [],
            )

            self.assertEqual(
                OrderedDict(
                    sorted(
                        EXPECTED_RESULT.items(),
                    ),
                ),
                OrderedDict(
                    sorted(
                        actual_result.items(),
                    ),
                ),
            )

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore,
    )
    def test_helper_functions(self):
        """Tests the helper functions used by the analyzer."""
        index_name = 'test'
        sketch_id = 1
        analyzer = safebrowsing.SafeBrowsingSketchPlugin(index_name, sketch_id)

        HELPERS = [
            self.check_sanitize_url,
            self.check_allowlist,
        ]

        for helper in HELPERS:
            helper(analyzer)

    def check_sanitize_url(self, analyzer):
        URLS = [
            ('http://w.com', 'http://w.com'),
            ('https://w.com', 'https://w.com'),
            ('Something before@https://w.com', 'https://w.com'),
            ('https://w.com and after', 'https://w.com'),
            ('nothing', ''),
        ]

        for entry, result in URLS:
            self.assertEqual(
                # pylint: disable=protected-access
                analyzer._sanitize_url(entry),
                result,
            )

    # pylint: disable=missing-docstring
    def check_allowlist(self, analyzer):
        ALLOW_LIST = [
            'lorem-*.com',
            'ipsum.dk',
            'dolo?.co.*',
        ]

        self.assertTrue(
            # pylint: disable=protected-access
            analyzer._is_url_allowlisted(
                'lorem-ipsum.com',
                ALLOW_LIST,
            ),
        )

        self.assertFalse(
            # pylint: disable=protected-access
            analyzer._is_url_allowlisted(
                'ipsum.com',
                ALLOW_LIST,
            ),
        )

        self.assertTrue(
            # pylint: disable=protected-access
            analyzer._is_url_allowlisted(
                'dolor.co.dk',
                ALLOW_LIST,
            ),
        )

        self.assertFalse(
            # pylint: disable=protected-access
            analyzer._is_url_allowlisted(
                'www.amet.com',
                ALLOW_LIST,
            ),
        )



"""Tests for DomainPlugin."""
from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers import domain
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestDomainPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    # Mock the Elasticsearch datastore.
    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_domain_analyzer_class(self):
        """Test core functionality of the analyzer class."""
        index_name = 'test'
        sketch_id = 1
        analyzer = domain.DomainSketchPlugin(index_name, sketch_id)
        self.assertEqual(analyzer.index_name, index_name)
        self.assertEqual(analyzer.sketch.id, sketch_id)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the annotation models."""

from __future__ import unicode_literals

from timesketch.lib.testlib import BaseTest


class AnnotationBaseTest(BaseTest):
    """Base class for common tests."""

    def _test_annotation(self, parent):
        """Test that the annotations has our test user as owner."""
        self.assertEqual(parent[0].user, self.user1)
        self.assertEqual(len(parent), 1)


class LabelModelTest(AnnotationBaseTest):
    """Test the label annotation."""

    def test_label_annotation(self):
        """Test that the label is associated with our test sketch."""
        self._test_annotation(self.sketch1.labels)
        # pylint: disable=unsubscriptable-object
        self.assertEqual(self.sketch1.labels[0].label, 'Test label')


class StatusModelTest(AnnotationBaseTest):
    """Test the status annotation."""

    def test_status_annotation(self):
        """Test that the status is associated with our test sketch."""
        self._test_annotation(self.sketch1.status)
        # pylint: disable=unsubscriptable-object
        self.assertEqual(self.sketch1.status[0].status, 'Test status')


class CommentModelTest(AnnotationBaseTest):
    """Test the comment annotation."""

    def test_comment_annotation(self):
        """Test that the comment is associated with our test event."""
        self._test_annotation(self.event.comments)
        # pylint: disable=unsubscriptable-object
        self.assertEqual(self.event.comments[0].comment, 'test')
